from .plots.deprecated import cluster_embedding, embedding_with_binding
from .plots import (
    barplot,
    hist_lvst_dist,
    length_distribution,
    logo_plot,
    protein_embedding,
    protein_network_embedding,
    rarefraction_curves,
    stacked_aa_distribution,
    levenshtein_clustering,
    sample_cluster,
    clone_fraction,
    diversity_plot,
    hist_lvst_dist_bind,
    multiple_length_plot,
    protein_embedding_umap,
)
from .plots.matrices import make_matrix
import matplotlib.pyplot as plt
from .augment_data.binding_data import collect_binding_data
from .augment_data.uploader import upload
import pandas as pd
import os
from .settings import change_settings, reports, plot_styler
import subprocess
from .augment_data.uploader import create_alignment_report
from .settings.general_instructions import print_instructions
from .plots.matrices.read_matrix import read_matrix
from .augment_data.mixcr_nils import check_mixcr
from .settings.aumotative_report import AumotativeReport
from .settings.figure import MyFigure, save_matrix
from .settings.markdown_builder import create_quarto
import warnings
from .settings.subplots_manager import Subplotter


class PlotManager:
    def __init__(
        self,
        experiment=None,
        test_version=False,
        length_threshold=6,
        min_read_count=3,
        no_automation=False,
        module_dir=None,
        allow_binding_data=True,
        remove_gaps=True,
        remove_errors=True,
        show_df = True,
    ):
        """
        :param experiment: the name of the experiment you want to analyse
        :param test_version: Some settings are different if the pipeline is launched in test mode
        :param length_threshold: Filter all amino acid sequences smaller than this value
        :param min_read_count: Filter all amino acid sequences with less reads than this value
        :param no_automation: If True, the pipeline will not launch the automatic plotting tool
        :param module_dir: If you want to use the pipeline in a different directory than the main directory, you can specify it here
        :param allow_binding_data: Default is True. If set to False, the pipeline will not ask for binding data
        :param remove_gaps: Default is True. If set to False, the pipeline will not remove sequences with gaps
        :param remove_errors: Default is True. If set to Flse, the pipeline will not remove sequences with errors (*).
        :param show_df: Default is True. If set to False, the pipeline will not show the dataframe in an interactive window. 
        """
        self.is_test = test_version
        self.Settings = change_settings.Settings()
        self.Settings.check_dirs()
        self.global_params = self.Settings.read_global_vars()
        self.remove_gaps = remove_gaps
        if module_dir:
            self.module_dir = module_dir
        else:
            self.module_dir = self.Settings.module_dir

        if experiment == None:
            self.sequencing_report, self.alignment_report, self.experiment = upload()
        else:
            self.sequencing_report = pd.read_csv(
                os.path.join(
                    self.module_dir,
                    "my_experiments",
                    experiment,
                    "sequencing_report.csv",
                )
            )
            self.alignment_report = create_alignment_report(self.module_dir, experiment)
            self.experiment = experiment
        self.region_string = self.global_params["region_of_interest"]
        if self.region_string == "":
            self.region_string = "CDR3"
        binding_report = reports.BindingReport(self.module_dir, self.experiment)
        if allow_binding_data != False:
            if os.path.isfile(allow_binding_data):
                self.binding_data = pd.read_csv(allow_binding_data)
            else:
                self.binding_data = binding_report.ask_binding_data()
        else:
            self.binding_data = None
        # check general settings and load them
        self.region_of_interest = "aaSeq" + self.region_string
        (
            self.plot_path,
            self.mixcr_plots_path,
            self.experiment_path,
            self.report_path,
        ) = self.Settings.check_dirs_automation(
            self.experiment, self.region_of_interest
        )
        self.font_settings = self.Settings.read_font_settings()
        self.legend_settings = self.Settings.read_legend_settings()
        self.colorbar_settings = self.Settings.read_colorbar_settings()
        # experiment_path = self.Settings.get_experiment_path(self.experiment)

        self.Report = reports.SequencingReport(self.sequencing_report)
        self.Report.check_sample_name(
            module_dir=self.module_dir, experiment_name=self.experiment
        )
        self.experiments_list = self.sequencing_report["Experiment"].unique().tolist()
        self.Report.prepare_seq_report(
            self.region_string,
            length_threshold=length_threshold,
            min_read_count=min_read_count,
            remove_gaps=self.remove_gaps,
            remove_errors=remove_errors,
        )
        # self.Report.map_exp_names(self.unique_experiments)
        self.avail_regions = self.Report.get_fragment()
        self.sequencing_report = self.Report.sequencing_report

        self.ControlFigure = MyFigure(test_version)
        self.ControlFigure.set_backend()
        self.subplot_list = []
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        # self.settings_saver = change_save_settings.Change_save_settings()
        self.preferred_sample = self.experiments_list[0]
        self.java_heap_size = 1000
        self.change_preferred_antigen()
        self.Automation = None
        if self.Settings.automation == True and no_automation == False:
            self.full_analysis()
        if self.is_test != True:
            print_instructions()
        if show_df == True:
            self.show_dataframe()

            
    def show_dataframe(self):
        """Shows you the main dataframe in an interactive window where you can also create your own plots and filter your data.
        """
        import importlib.util
        if importlib.util.find_spec("pandasgui"):
            if self.binding_data is not None:
                report = self.merge_bind_seq_report()
            else:
                report = self.sequencing_report
            from pandasgui import show
            show(report)
        else:
            print("Please install pandasgui to use this functionality. You can use pip install pandasgui.")
            return
        

    def create_report(self):
        self.Settings.move_markdown_files()
        print("Install quarto from: https://quarto.org/docs/get-started/")
        print(
            "You also might to install Jupyter. In general, I recommend to use VS code for this purpose."
        )
        print(
            f"This function should only be called after the initial automation, since it relies on the file structure in {self.plot_path}\nAlso do not change the names of the directories or files"
        )
        print("Test Quarto")
        subprocess.run(["quarto", "check"])
        print("Render qmd file")
        create_quarto(
            self.experiment, self.plot_path, self.binding_data, self.experiments_list
        )
        subprocess.run(
            ["quarto", "render", os.path.join(self.plot_path, self.experiment + ".qmd")]
        )
        print(
            f"You can find the html file to the report at: {self.plot_path}/{self.experiment}.qmd"
        )

    def add_to_subplot(self, figure_name="No.1", capture="", dir="tmp_quarto"):
        """You can add the current plot to a certain figure with the given figure_name and quarto will generate an html document with all subplots which have been added until that point.\nYou need to install quarto to use this feature.

        Args:
            figure_name (str, optional): The name of the figure. If you change it you will add your plots to another figure and a new class instance starting with subplot_{figure_name} is created. Defaults to "No.1".
        """
        if hasattr(self, f"subplot_{figure_name}"):
            pass
        else:
            setattr(
                self, f"subplot_{figure_name}", Subplotter(figure_title=figure_name)
            )
            self.subplot_list.append(getattr(self, f"subplot_{figure_name}"))
        subplot_instance = getattr(self, f"subplot_{figure_name}")
        subplot_instance.add_as_subplot(self.ControlFigure.fig, capture, dir)

    def show_subplot(self, figure_name="No.1", dir="tmp_quarto", captures=None):
        print("Test Quarto")
        subprocess.run(["quarto", "check"])
        if hasattr(self, f"subplot_{figure_name}"):
            subplot_instance = getattr(self, f"subplot_{figure_name}")
        else:
            setattr(
                self, f"subplot_{figure_name}", Subplotter(figure_title=figure_name)
            )
            subplot_instance = getattr(self, f"subplot_{figure_name}")
            subplot_instance.update_files(dir)
        if captures != None:
            subplot_instance.captures = []
            for capture in captures:
                subplot_instance.update_capture(capture)
        subplot_instance.make_figure(dir=dir)

    def change_java_heap_size(self, new_size):
        self.java_heap_size = new_size

    def export_sequencing_report(self):
        self.sequencing_report.to_csv(
            os.path.join(self.experiment_path, "sequencing_report_processed.csv")
        )
        print(
            f"Lowest peptide sequence length: {self.sequencing_report['aaSeqCDR3'].str.len().min()}"
        )
        print(f"lowest read count: {self.sequencing_report['readCount'].min()}")

    def chat(
        self,
    ):
        """
        :return: starts a conversation with your data. Be creative! You can ask any question and even create plots :)
        """
        if self.Automation == None:
            if self.binding_data is not None:
                merged_reprot = self.merge_bind_seq_report()
                print("Your data looks like: \n")
                merged_reprot.head(10)
                self.Automation = AumotativeReport(
                    merged_reprot, self.Report.origin_seq_report, self.global_params
                )
            else:
                print("Your data looks like: \n")
                self.sequencing_report.head(10)
                self.Automation = AumotativeReport(
                    self.sequencing_report,
                    self.Report.origin_seq_report,
                    self.global_params,
                )

        response = self.Automation.chat_()
        return response

    def save_in_plots(self, enter_filename):
        plt.savefig(
            fname=os.path.join(self.plot_path, enter_filename + f".png"),
            dpi=300,
            format="png",
            bbox_inches="tight",
        )

    def get_best_binder(self):
        if self.binding_data is None:
            return None
        found_antigens = self.binding_data.columns.to_list()[1:]
        best_binding = {}

        for i in self.experiments_list:
            sub_report = self.sequencing_report[
                self.sequencing_report["Experiment"] == i
            ]
            merged_report = sub_report.merge(self.binding_data, how="inner")
            clone_fractions = merged_report["cloneFraction"]
            clone_fractions.fillna(0, inplace=True)
            best_binding_local = {}
            for antigen in found_antigens:
                binding_values = merged_report[antigen]
                binding_values.fillna(0, inplace=True)
                significance_binding = (binding_values * clone_fractions).mean()
                if antigen in best_binding_local.keys():
                    if best_binding_local.get(antigen) < significance_binding:
                        best_binding_local[antigen] = significance_binding
                    else:
                        pass
                else:
                    best_binding_local[antigen] = significance_binding

            sorted_keys = sorted(
                best_binding_local, key=best_binding_local.get, reverse=True
            )

            # Get the best and second best keys
            best_key = sorted_keys[0]
            if len(found_antigens) >= 3:
                second_best_key = sorted_keys[1] if len(sorted_keys) > 1 else None
                best_binding[i] = [best_key, second_best_key]
            else:
                best_binding[i] = [best_key]
        return best_binding

    def model_protein():
        print(
            "Visit one of these pages and explore modelling proteins for free!"
            f"ESM FOLD: https://colab.research.google.com/github/sokrypton/ColabFold/blob/main/ESMFold.ipynb#scrollTo=CcyNpAvhTX6q"
        )

    def full_analysis(self):
        """
        :return: returns some plots about quality in /my_experiments/YOUR_EXPERIMENT_NAME/plots. Is called automatically when launching the pipeline.
        """
        (
            self.plot_path,
            self.mixcr_plots_path,
            self.experiment_path,
            self.report_path,
        ) = self.Settings.check_dirs_automation(
            self.experiment, self.region_of_interest
        )
        self.plot_path = os.path.join(
            self.module_dir,
            "my_experiments",
            self.experiment,
            "plots",
            self.region_of_interest,
        )
        report_seq_cluster = os.path.join(self.plot_path, "sequence_cluster", "reports")
        jaccard_report = os.path.join(self.report_path, "jaccard_identity" + ".xlsx")
        if not os.path.isdir(self.plot_path):
            os.makedirs(self.plot_path)
        if not os.path.isdir(os.path.join(self.plot_path, "length_distributions")):
            os.makedirs(os.path.join(self.plot_path, "length_distributions"))

        if not os.path.isdir(os.path.join(self.plot_path, "clone_fraction")):
            os.makedirs(os.path.join(self.plot_path, "clone_fraction"))
        if not os.path.isdir(os.path.join(self.plot_path, "rarefraction_curves")):
            os.makedirs(os.path.join(self.plot_path, "rarefraction_curves"))
        if not os.path.isdir(os.path.join(self.plot_path, "logo_plots")):
            os.makedirs(os.path.join(self.plot_path, "logo_plots"))
        if not os.path.isdir(os.path.join(self.plot_path, "sequence_cluster")):
            os.makedirs(os.path.join(self.plot_path, "sequence_cluster"))

        if not os.path.isdir(os.path.join(self.plot_path, "sequence_embedding")):
            os.makedirs(os.path.join(self.plot_path, "sequence_embedding"))

        if not os.path.isdir(os.path.join(self.plot_path, "sequence_embedding", "sgt")):
            os.makedirs(os.path.join(self.plot_path, "sequence_embedding", "sgt"))

        if not os.path.isdir(
            os.path.join(self.plot_path, "sequence_embedding", "protbert")
        ):
            os.makedirs(os.path.join(self.plot_path, "sequence_embedding", "protbert"))

        if not os.path.isdir(os.path.join(self.plot_path, "sequence_embedding", "T5")):
            os.makedirs(os.path.join(self.plot_path, "sequence_embedding", "T5"))

        if not os.path.isdir(report_seq_cluster):
            os.makedirs(report_seq_cluster)

        # plot
        try:

            mh_save_path = os.path.join(
                self.report_path, "morosita_horn_identity" + ".xlsx"
            )
            if not os.path.isfile(os.path.join(self.plot_path, "morosita_horn.png")):
                print("Start preparing morosita horn matrix.")
                self.morosita_horn(matrix_save_path=mh_save_path)
                self.save_in_plots("morosita_horn")
        except:
            print("Creating Morosita Horn matrix failed")
        # plot
        try:
            if not os.path.isfile(os.path.join(self.plot_path, "sorensen.png")):
                print("Start preparing sorensen matrix.")

                sorensen_save_path = os.path.join(
                    self.report_path, "sorensen_identity" + ".xlsx"
                )
                self.sorensen(matrix_save_path=sorensen_save_path)
                self.save_in_plots("sorensen")
        except:
            print("Creating sorense matrix failed")
        # plot
        try:
            if not os.path.isfile(os.path.join(self.plot_path, "jaccard.png")):
                print("Start preparing jaccard matrix.")

                jaccard_save_path = os.path.join(
                    self.report_path, "jaccard_identity" + ".xlsx"
                )
                self.jaccard(matrix_save_path=jaccard_save_path)
                self.save_in_plots("jaccard")
        except:
            print("Creating sorense matrix failed")

        # clone fract plot
        for experiment in self.experiments_list:
            try:
                self.rel_seq_abundance(experiment)
                self.save_in_plots(os.path.join("clone_fraction", experiment))
            except:
                print(f"Clone fraction for {experiment} failed")
        # diversity plot

        for method in ["Shannon", "InverseSimpson"]:
            try:
                self.sample_diversity(method)
                self.save_in_plots(os.path.join("diversity_" + method))
            except:
                print(f"Diversity for {method} plot failed")
        # plot
        for experiment in self.experiments_list:
            try:
                self.lengthDistribution_single(experiment)
                self.save_in_plots(os.path.join("length_distributions", experiment))
            except:
                print(f"Length Distribution for {experiment} failed")
        # mixcr plots
        try:
            print(
                "Begin with creating mixcr plots. You can find a description about them here:\n https://mixcr.com/mixcr/reference/mixcr-exportPlots/"
            )
            self.export_mixcr_quality()
            self.mixcr_explain_diversity()
            self.mixcr_vdj_usage()
        #  self.mixcr_overlap()
        except:
            print(
                f".clns files could not be found. Most certainly, you have uploaded the sequencing report."
            )
        # plot
        try:
            self.alignment_quality()
            self.save_in_plots("alignment_quality")
        except:
            print("Alignment reports could not be found")

        # plot
        try:
            if not os.path.isfile(os.path.join(self.plot_path, "rarefraction_all.png")):
                print("create rarefraction curves of all samples")
                self.rarefraction_curves(self.experiments_list)
                self.ControlFigure.fig.tight_layout()
                self.save_in_plots("rarefraction_all")
        except:
            print("creating rarefraction plot of all samples failed")

        if not os.path.isdir(os.path.join(self.plot_path, "rarefraction_curves")):
            os.mkdir(os.path.join(self.plot_path, "rarefraction_curves"))

        for i in self.experiments_list:
            if not os.path.isfile(
                os.path.join("rarefraction_curves", "rarefraction_curves_" + i + ".png")
            ):
                try:
                    self.rarefraction_curves(samples=[i])
                    self.save_in_plots(
                        os.path.join("rarefraction_curves", "rarefraction_curves_" + i)
                    )
                except:
                    pass

        # plot
        for experiment in self.experiments_list:
            try:
                if not os.path.isfile(
                    os.path.join(self.plot_path, "logo_plots", experiment + ".png")
                ):
                    self.logoPlot(sample=experiment)
                    self.save_in_plots(os.path.join("logo_plots", experiment))
            except:
                print(f"Logo Plot for {experiment} failed")

        print(
            "Cluster NGS sequences in dendrograms and network plots using levenshtein distance of 2 for network plots and ls-distance of 1 for dendrogram. Batch size is set to 1000."
        )
        # plot
        try:
            print(
                f"Create clustering for all samples for top 50% of clone fraction and for maximum of 100 reads"
            )
            self.connect_samples()
            self.save_in_plots(os.path.join(self.plot_path, "ls_connection_all"))
        except:
            print("Creation of cluster for all samples failed.")
        for single_experiment in self.experiments_list:
            try:
                if not os.path.isfile(
                    os.path.join(
                        self.plot_path,
                        "sequence_cluster",
                        single_experiment + "ls_dendro.png",
                    )
                ):
                    self.levenshtein_dendrogram(
                        sample=single_experiment, max_cluster_dist=1
                    )
                    self.save_in_plots(
                        os.path.join(
                            "sequence_cluster", single_experiment + "ls_dendro"
                        )
                    )
                if not os.path.isfile(
                    os.path.join(
                        self.plot_path,
                        "sequence_cluster",
                        single_experiment + "ls_cluster.png",
                    )
                ):
                    self.basic_cluster(
                        samples=[single_experiment],
                        max_ld=2,
                        batch_size=1000,
                        label_type=None,
                        save_report_path=os.path.join(
                            report_seq_cluster,
                            f"{single_experiment}" + "ls_cluster" + ".xlsx",
                        ),
                    )
                    self.save_in_plots(
                        os.path.join(
                            "sequence_cluster", single_experiment + "ls_cluster"
                        )
                    )
            except:
                print(f"Sequence clustering failed at: {single_experiment}")
        threshold_identity = 0.2

        while True:
            overlapping_samples = read_matrix(
                threshold=threshold_identity, path_matrix=mh_save_path
            )
            # if None path_matrix does not exist
            if overlapping_samples == None:
                if mh_save_path == jaccard_report:
                    overlapping_samples = None
                    break
                else:
                    mh_save_path = jaccard_report
                    continue
            else:
                mean_length = sum(len(v) for v in overlapping_samples.values()) / len(
                    overlapping_samples
                )
                if mean_length > 2.5:
                    break
                elif threshold_identity < 0.05:
                    break
                else:
                    threshold_identity -= 0.01

        print(
            "Start sequence embedding of samples with 1000 iterations for tsne and a batch size of 300"
        )
        if overlapping_samples != None:
            for single_experiment in list(overlapping_samples.keys()):
                if not os.path.isfile(
                    os.path.join(
                        self.plot_path,
                        "sequence_embedding",
                        "sgt",
                        single_experiment + "embedding_tsne.png",
                    )
                ):
                    try:
                        self.embedding_tsne(
                            samples=overlapping_samples[single_experiment],
                            model="sgt",
                            strands=False,
                            batch_size=300,
                            iterations_tsne=1000,
                        )
                        self.save_in_plots(
                            os.path.join(
                                "sequence_embedding",
                                "sgt",
                                single_experiment + "embedding_tsne",
                            )
                        )
                    except:
                        print(f"Clustering with TSNE failed for {single_experiment}.")

        if overlapping_samples != None:
            for single_experiment in list(overlapping_samples.keys()):
                if not os.path.isfile(
                    os.path.join(
                        self.plot_path,
                        "sequence_embedding",
                        "T5",
                        single_experiment + "embedding.png",
                    )
                ):
                    try:
                        self.embedding_tsne(
                            samples=overlapping_samples[single_experiment],
                            strands=False,
                            batch_size=300,
                        )
                        self.save_in_plots(
                            os.path.join(
                                "sequence_embedding",
                                "T5",
                                single_experiment + "embedding_tsne",
                            )
                        )
                    except:
                        print(f"Clustering with TSNE failed for {single_experiment}.")

        else:
            print(
                "Neither Morosita Horn nor Jaccard matrix was generated, thus no sequence embedding can be created."
            )
        if self.region_of_interest not in self.binding_data.columns.tolist():
            print(f"There is no binding data available for {self.region_of_interest}")
            best_binder = None
        else:
            best_binder = self.get_best_binder()

        if best_binder == None:
            print("No binding data was found. Thus, no binding plots can be created.")

        else:
            clustering_antigens_path = os.path.join(
                self.plot_path, "clustering_antigens"
            )
            if not os.path.isdir(clustering_antigens_path):
                os.mkdir(clustering_antigens_path)
            experiment_keys = list(best_binder.keys())
            print(
                "Starting clustering of binding data for antigens and sequences using PCA and t-SNE.\nPerplexity is set to 25.\n70 principal components are taken for t-SNE.\n1000 iterations are used for tsne. Batch size is set to 300."
            )
            report_tsne_cluster = os.path.join(
                self.plot_path, "clustering_antigens", "reports"
            )
            if not os.path.isdir(report_tsne_cluster):
                os.mkdir(report_tsne_cluster)

            # plot
            for experiment in experiment_keys:
                try:
                    if not os.path.isfile(
                        os.path.join(
                            self.plot_path,
                            "clustering_antigens",
                            experiment + "tsne_cluster_AG.png",
                        )
                    ):
                        self.cluster_binding_data(
                            samples=[experiment],
                            batch_size=300,
                            antigens=best_binder[experiment],
                            show_antigen_names=False,
                            iterations_tsne=1000,
                            save_report_path=os.path.join(
                                report_tsne_cluster,
                                experiment + f"_best_binder{experiment}" + ".xlsx",
                            ),
                        )
                        self.save_in_plots(
                            os.path.join(
                                "clustering_antigens", experiment + "tsne_cluster_AG"
                            )
                        )
                except:
                    print(
                        f"Cluster based on sequence embedding combined with binding data failed for: {experiment}"
                    )
            dendro_path = os.path.join(
                self.plot_path, "clustering_antigens", "dendro_binding"
            )

            if not os.path.isdir(dendro_path):
                os.mkdir(dendro_path)
            ls_cluster_path = os.path.join(
                self.plot_path, "clustering_antigens", "ls_binding_cluster"
            )
            if not os.path.isdir(ls_cluster_path):
                os.mkdir(ls_cluster_path)
            print(
                "Create dendrogram with binding data and levenshtein distance of 2 and batch size of 300 for the dendrogram and 1000 for the network plots."
            )

            report_ls_cluster = os.path.join(
                self.plot_path, "clustering_antigens", "ls_binding_cluster", "reports"
            )
            if not os.path.isdir(report_ls_cluster):
                os.mkdir(report_ls_cluster)
            # plot
            for experiment in experiment_keys:
                try:
                    plt.close("all")
                    fig = self.dendro_bind(
                        sample=experiment,
                        antigens=best_binder[experiment],
                        batch_size=300,
                    )
                    if fig == False:
                        continue
                    else:
                        plt.close(fig)
                        self.save_in_plots(
                            os.path.join(
                                "clustering_antigens",
                                "dendro_binding",
                                experiment + "cluster_dendrogram",
                            )
                        )
                        try:
                            plt.close("all")
                            self.cluster_one_AG(
                                antigen=best_binder[experiment][0],
                                specific_experiments=[experiment],
                                batch_size=1000,
                                max_ld=2,
                                preferred_cmap="Reds",
                                label_type=None,
                                save_report_path=os.path.join(
                                    report_ls_cluster,
                                    experiment + f"_ls_binding_cluster" + ".xlsx",
                                ),
                            )
                            self.save_in_plots(
                                os.path.join(
                                    "clustering_antigens",
                                    "ls_binding_cluster",
                                    experiment + f"_ls_binding_cluster",
                                )
                            )
                        except:
                            print(
                                "Could not create levenshtein distance cluster for best binder"
                            )
                except:
                    print(f"Dendrogram with binding data failed for {experiment}")

        print(f"The pipeline has created some plots in: {self.plot_path}")

        print(
            "You can create and automatic report of the generated analysis by installing Quarto(https://quarto.org/docs/get-started/) and type: plot.create_report()"
        )

    def change_preferred_antigen(self, antigen=None):
        if self.binding_data is not None:
            columns_bind_data = self.binding_data.columns.to_list()
            if antigen == None:
                antigen = columns_bind_data[1]
            self.preferred_antigen = antigen
        else:
            self.preferred_antigen = None

    def change_preferred_sample(self, sample):
        assert (
            sample in self.experiments_list
        ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.preferred_sample = sample

    def change_filter(
        self,
        length_threshold_aa=6,
        min_read_count=2,
        remove_gaps=True,
        remove_errors=True,
    ):
        """
        length_threhsold_aa: The minimum sequence length the sequence should at least have.
        min_read_count (int): The minimum read count the sequences should at least have.
        remove_gaps (bool): You can decide whether you want to discard sequences which have a gap or are not divisible by 3.
        remove_errors (bool): You can decide whether you want to discard sequences which have sequencing errors (*).
        """
        self.Report.prepare_seq_report(
            self.region_string,
            length_threshold_aa,
            min_read_count,
            remove_gaps,
            remove_errors,
        )
        self.sequencing_report = self.Report.sequencing_report

    def change_region(
        self,
        region=None,
        remove_gaps=True,
        remove_errors=True,
        length_threshold_aa=6,
        min_read_count=3,
    ):
        """
        region (str): The region of interest you would like to analyse
        remove_gaps (bool): You can decide whether you want to discard sequences which have a gap or are not divisible by 3.
        remove_errors (bool): You can decide whether you want to discard sequences which have sequencing errors (*).
        length_threhsold_aa: The minimum sequence length the sequence should at least have.
        min_read_count (int): The minimum read count the sequences should at least have.
        :return: changes the region you want to analyse
        """

        intermediate = self.region_of_interest
        possible_regions = self.avail_regions
        if region == None:
            region_string = input(
                f"Which region do you want to plot? ExpoSeq could find the following regions: {self.avail_regions}."
            )
        else:
            region_string = region
        self.remove_gaps = remove_gaps
        if region_string in possible_regions:
            if not region_string == "targetSequences":
                self.region_string = region_string.replace("nSeq", "")
                self.Report.prepare_seq_report(
                    self.region_string,
                    length_threshold=length_threshold_aa,
                    min_read_count=min_read_count,
                    remove_gaps=self.remove_gaps,
                    remove_errors=remove_errors,
                )
            else:
                self.region_string = "targetSequences"
                self.Report.prepare_seq_report(
                    self.region_string,
                    length_threshold=length_threshold_aa,
                    min_read_count=min_read_count,
                    remove_gaps=self.remove_gaps,
                    remove_errors=remove_errors,
                )
            self.sequencing_report = self.Report.sequencing_report
            self.region_of_interest = "aaSeq" + self.region_string
        else:
            print(
                f"The region you want to plot is not valid. The options are: {self.avail_regions}"
            )
            self.region_of_interest = intermediate
        plot_dir = os.path.dirname(self.plot_path)
        self.plot_path = os.path.join(plot_dir, self.region_of_interest)

    def discard_samples(self, samples_to_discard):
        """

        :param samples_to_discard: the name of the samples in a list which you want to delete from your further analysis
        :return: updates the sequencing report where the values for the given samples do not exist anymore

        """
        assert (
            type(samples_to_discard) == list
        ), "You have to give a list with the samples you want to discard"
        self.sequencing_report = self.sequencing_report[
            ~self.sequencing_report["Experiment"].isin(samples_to_discard)
        ]

    def add_binding_data(self):
        """ "
        :return: adds binding data to your analysis. You can add mutliple files with the filechooser or a given path to the file. For more information about the necessary file structure, check the supplementary information.
        """
        self.binding_data = collect_binding_data(binding_data=self.binding_data)

        self.change_preferred_antigen(self.binding_data.columns.to_list()[1])

        # self.binding_data.to_csv("binding_data.csv")

    def merge_bind_seq_report(self, merge_technique="left"):
        """
        :param: merge_technique: default "left". You can choose between "left", "right", "inner" and "outer" for the merge technique.
        :return: a python variable which contains the merged sequencing report and binding report which can be saved with report.to_csv("filename.csv")
        """
        assert (
            self.binding_data is not None
        ), "You have not given binding data. You can add it with the add_binding_data function"

        merged_reports = pd.merge(
            self.sequencing_report,
            self.binding_data,
            on="aaSeqCDR3",
            how=merge_technique,
        )
        merged_reports.fillna(0, inplace=True)
        return merged_reports

    def print_antigens(self):
        """
        :return: prints the antigens (columns) of your binding data
        """
        try:
            print(self.binding_data.columns.to_list()[1:])
        except:
            print("You have not given binding data.")

    def print_samples(self):
        """
        :return: prints the names of your samples, so you can insert them in lists or similar for the analysis with some plots
        """
        print(self.experiments_list)

    def save(self, enter_filename, dpi=300, format="png"):
        """
        :return: can be used to save your plots. It will ask you automatically for the directory where you want to save it
        """
        print(
            "Choose the directory where you want to save your plot with the filechooser"
        )
        try:
            from tkinter import filedialog

            save_dir = filedialog.askdirectory()
        except:
            save_dir = ""

        plt.savefig(
            fname=os.path.join(save_dir, enter_filename + f".{format}"),
            dpi=dpi,
            format=format,
        )

    def close(self):
        plt.close()

    def change_experiment_names(self, specific=None):
        """
        :param specific: optional parameter. You can use this function to change the names of a specific sample.
        :return:You can use this function to change the name of your samples. Thus, you can change the labels of your plots.
        """

        self.expreiments_list = self.sequencing_report["Experiment"].unique().tolist()
        experiments_dic = dict(zip(self.experiments_list, self.experiments_list))
        if specific != None:
            assert type(specific) == str, (
                "You have to give a string (inside: "
                ") as input for the specific sample you want to change"
            )

        if specific == None:
            print(
                "You will no go through all your samples and you will be able to change their names. If you want to skip a samples, just press enter."
            )
            for key in self.experiments_list:
                print(f"Current value for {key}: {experiments_dic[key]}")
                new_value = input("Enter a new value or press any key to skip")
                if len(new_value) > 1:
                    experiments_dic[key] = new_value
                else:
                    pass
        else:
            new_value = input("Enter the new name for " + specific)
            experiments_dic[specific] = new_value
        self.Report.renew_exp_names_origin(experiments_dic, self.experiment_path)
        self.experiments_list = list(experiments_dic.values())
        self.sequencing_report["Experiment"] = self.sequencing_report["Experiment"].map(
            experiments_dic
        )

    def alignment_quality(self, log_transformation=False):
        """

        :param log_transformation: optional parameter. You can set it to True if you want to log transform your data
        :return: Creates a figure where you can see the Overall Reads per sample and the number of reads which could be aligned.
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert log_transformation in [
            True,
            False,
        ], "You have to give True or False as input for the log transformation"
        self.ControlFigure.clear_fig()
        try:
            barplot.barplot(
                self.ControlFigure.ax,
                self.alignment_report,
                self.sequencing_report,
                self.font_settings,
                self.legend_settings,
                apply_log=log_transformation,
            )

            self.ControlFigure.update_plot()
            self.style = plot_styler.PlotStyle(
                self.ControlFigure.ax, self.ControlFigure.plot_type
            )
        except:
            print("Currenlty you do not have loaded an alignment report.")

    def aa_distribution(self, sample=None, region=[3, 7], protein=True):
        """
        :param sample: The sample you would like to analyze
        :param region: The input is a list and you can specify with the first value the beginning region and with the second value the end region within your sequences for which you want to know the amino acid composition. For instance: [3,7]
        :param protein: Default True. If you would like to analyze nucleotide sequences, set it to False
        :return: Returns a plot which shows the amino acid distribution in the given sequence range.
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert (
            sample in self.experiments_list
        ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert type(sample), "You have to give a string as input for the sample"
        assert type(
            region
        ), "You have to give a list with the start and end position of the region you want to analyze. For instance: [3,7]"
        assert protein in [
            True,
            False,
        ], "You have to give True or False as input for the protein parameter"
        self.ControlFigure.clear_fig()
        stacked_aa_distribution.StackedAADistribution(
            self.sequencing_report,
            sample,
            region,
            self.region_of_interest,
            protein,
            self.font_settings,
            self.ControlFigure.ax,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def rarefraction_curves(self, samples=None):
        """
        :param samples: you insert a list which contains the sample names
        :return: Shows your the rarefraction curves for the given samples.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert (
            not incorrect_samples
        ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if type(samples) != list:
            print(
                "You have to give the sample names in the list, also if it is only one! A list is a container which is marked through:[] . Please try again."
            )
        else:
            self.ControlFigure.clear_fig()
            rarefraction_curves.RarefractionCurves(
                sequencing_report=self.sequencing_report,
                samples=samples,
                region_of_interest=self.region_of_interest,
                ax=self.ControlFigure.ax,
                font_settings=self.font_settings,
                legend_settings=self.legend_settings,
            )

            self.ControlFigure.update_plot()
            self.style = plot_styler.PlotStyle(
                self.ControlFigure.ax, self.ControlFigure.plot_type
            )

    def logoPlot(
        self,
        sample=None,
        highlight_specific_pos=None,
        method="proportion",
        chosen_seq_length=None,
        color_scheme="chemistry",
        **kwargs,
    ):
        """
        :param sample: Insert the sample names as list. You can analyse multiple or only one.
        :param highlight_specific_pos: optional. you can highlight a specific position. For instance if you want to highlight the 3rd position, you insert 3.
        :param method: You can specify whether you want to have on your y axis the frequency of the amino acids or the information content in bits. The default is proportion. If you want to have the information content, insert "bits"
        :param chosen_seq_length: Max per default. If you plot the length distribution it will take the sequences with the highest bar there. You always analyze only one sequence length! You can enter an integer to analyze other lengths.
        :param color_scheme: Default is skylign_protein. You can choose between skylign_protein, chemistry, weblogo_protein, hydrophobicity, dmslogo_charge, charge, NajafabadiEtAl2017.
        :return: A logo Plot which shows you the composition of aminoacids per position
        """
        logo_schemes = [
            "skylign_protein",
            "chemistry",
            "weblogo_protein",
            "hydrophobicity",
            "dmslogo_charge",
            "charge",
            "NajafabadiEtAl2017",
        ]
        assert (
            color_scheme in logo_schemes
        ), f"You have to give a valid color scheme. The options are: {logo_schemes}"
        if sample == None:
            sample = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if highlight_specific_pos != None:
            assert (
                type(highlight_specific_pos) == int
            ), "You have to give an integer as input for the specific position you want to highlight"

        self.ControlFigure.clear_fig()

        logo_plot.LogoPlot(
            self.ControlFigure.ax,
            self.sequencing_report,
            self.region_of_interest,
            sample,
            highlight_specific_pos,
            self.font_settings,
            chosen_seq_length,
            method,
            color_scheme,
            self.avail_regions,
            **kwargs,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def logoPlot_multi(
        self,
        samples="all",
        chosen_seq_length=1,
        method="proportion",
    ):
        """
        :param samples: You analyze all samples per default. If you want to analyze specific samples it has to be a list with the corresponding sample names
        :param chosen_seq_length: 16 per default. You always analyze online one sequence length! You can change it if you would like
        :return: Gives you in one figure one logoPlot per sample.
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        if samples != "all":
            assert (
                type(samples) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in samples if x not in self.experiments_list]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        if chosen_seq_length != "max":
            assert (
                type(chosen_seq_length) == int
            ), "You have to give an integer as input for the sequence length you want to analyze"
        self.ControlFigure.clear_fig()

        logo_plot.plot_logo_multi(
            self.ControlFigure.fig,
            self.sequencing_report,
            samples,
            self.font_settings,
            self.region_of_interest,
            method,
            chosen_seq_length,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    #  self.ControlFigure.tighten()

    def lengthDistribution_single(self, sample=None):
        """

        :param sample: name of the sample from which you would like to see the length distribution
        :return: Shows you the length Distribution of your sequences of the given sample
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert (
            sample in self.experiments_list
        ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        length_distribution.LengthDistributionSingle(
            self.sequencing_report,
            sample,
            self.region_of_interest,
            self.ControlFigure.ax,
            self.font_settings,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def lengthDistribution_multi(self, samples="all"):
        """
        :param samples: You analyze all samples per default. If you want to analyze specific samples it has to be a list with the corresponding sample names
        :return: Outputs one figure with one subplot per sample which shows you the distribution of the sequence length
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        if samples != "all":
            assert (
                type(samples) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [x for x in samples if x not in self.experiments_list]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        length_distribution.length_distribution_multi(
            self.ControlFigure.fig,
            self.sequencing_report,
            samples,
            self.font_settings,
            self.region_of_interest,
            test_version=self.is_test,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    #  self.ControlFigure.tighten()

    def length_distribution_all(self, plot_type="boxplot"):
        """Returns either a box or a violin plot which shows the distribution of the nucleotide sequence length.

        Args:
            plot_type (str, optional): This parameter controls if you want to have the length distribution visualized as box or violin plot. Type violin to show the Violinplot. Defaults to "boxplot".
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        multiple_length_plot.LengthPlotMultiple(
            sequencing_report=self.sequencing_report,
            region_of_interest=self.region_of_interest,
            ax=self.ControlFigure.ax,
            font_settings=self.font_settings,
            plot_type=plot_type,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def rel_seq_abundance(
        self,
        sample=None,
        visualize_sequences=True,
        prefered_cmap="Reds",
        top_clone_fraction=0.75,
        seqs_viz_fraction=0.3,
        alpha_val=0.75,
        pad_rectangles=True,
        force_reducing=500,
        limit_seq_filter=10,
    ):
        """
        :param samples: For a qualitative analysis choose samples from the same panning experiment. Input is a list
        :param visualize_sequences: Default is True. If you want to visualize a fraction of the sequences in the rectangles, set it to True
        :param prefered_cmap: Default is Reds. You can choose any colormap from matplotlib
        :param top_clone_fraction: Default is 0.95. You can choose the top clone fraction you want to visualize
        :param seqs_viz_fraction: Default is 0.75. You can choose the fraction of sequences you want to visualize
        :param alpha_val: Default is 0.5. You can choose the alpha value of the rectangles
        :param pad_rectangles: Default is True. If you want to have a padding between the rectangles, set it to True
        :param force_reducing: Default is 1000. If you have more than 1000 sequences after filtering by taking the top 95 % clones, this threshold automatically limits the sequences to 1000. You can set it to None if you want to avoid this filtering.
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        assert (
            type(sample) == str
        ), "You have to give a list with the samples you want to analyze"
        assert (
            type(visualize_sequences) == bool
        ), "You have to give True or False as input for the visualize_sequences parameter"
        assert (
            type(prefered_cmap) == str
        ), "You have to give a string as input for the prefered_cmap parameter"
        assert (
            type(top_clone_fraction) == float
        ), "You have to give a float as input for the top_clone_fraction parameter"
        assert top_clone_fraction < 1, "The top_clone_fraction has to be smaller than 1"
        assert top_clone_fraction > 0, "The top_clone_fraction has to be larger than 0"
        assert (
            type(seqs_viz_fraction) == float
        ), "You have to give a float as input for the seqs_viz_fraction parameter"
        assert seqs_viz_fraction < 1, "The seqs_viz_fraction has to be smaller than 1"
        assert seqs_viz_fraction > 0, "The seqs_viz_fraction has to be larger than 0"
        assert (
            type(alpha_val) == float
        ), "You have to give a float as input for the alpha_val parameter"
        assert alpha_val < 1, "The alpha_val has to be smaller than 1"
        assert alpha_val > 0, "The alpha_val has to be larger than 0"
        assert (
            type(pad_rectangles) == bool
        ), "You have to give True or False as input for the pad_rectangles parameter"
        self.ControlFigure.clear_fig()
        clone_fraction.VisFrac(
            self.sequencing_report,
            self.ControlFigure.ax,
            sample,
            visualize_sequences,
            prefered_cmap,
            top_clone_fraction,
            seqs_viz_fraction,
            alpha_val,
            pad_rectangles,
            force_reducing,
            limit_seq_filter,
            self.font_settings,
            self.region_of_interest,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def save_cluster_report(self, cluster_report, path=None):
        if path == None:
            while True:
                cluster_report_save = input(
                    "Do you want to save the generated data? (Y/n)"
                )
                if cluster_report_save in ["Y", "y", "N", "n"]:
                    break
                else:
                    print("Please enter Y or n")
            if cluster_report_save.lower() in ["Y", "y"]:
                while True:
                    filename_matrix = input(
                        f"Enter a name for the file. The file will be saved in {self.report_path} in your IDE."
                    )
                    if filename_matrix.endswith(".xlsx"):
                        pass
                    else:
                        filename_matrix = filename_matrix + ".xlsx"
                    if not os.path.isfile(filename_matrix):
                        path = os.path.join(self.report_path, filename_matrix)
                        cluster_report.to_excel(path)
                        break
                    else:
                        print("This file already exists. Please choose another name.")
        else:
            cluster_report.to_excel(path)

    def umap_sample_cluster(
        self,
        samples=None,
        n_neighbors=20,
        min_dist=0.01,
        random_seed=42,
        densmap=True,
        metric="euclidean",
        model_choice="Rostlab/prot_t5_xl_half_uniref50-enc",
        show_strands=False,
        clone_size_factor=300,
        batch_size=1000,
        pca_components=50,
        prefered_cmap="viridis",
        save_report_path=None,
        iterations_umap = 1000,
    ):
        """You can use this plot to solely compare samples to each other with UMAP as the main dimension reduction technique.

        Args:
            samples (str, optional): Here you can insert a list of samples you would like to compare and cluster. Defaults to None.
            n_neighbors (int, optional): Here you can choose the numbers of neighbors a point has.Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should be between 5 to 50. Defaults to 20.
            min_dist (float, optional): controls how tightly the points will be set to each other. It should be the minimum distance points are allowed to be apart from each other in the low dimensional representation. Defaults to 0.01.
            random_seed (int, optional): Set a certain seed for reprodubility. Defaults to 42.
            densmap (bool, optional): This parameter allows you to visualize points more densily which are also more dense in all dimensions to each other.. Defaults to True.
            metric (str, optional): You need to insert a string as input which is the distance metric for the UMAP algorithm. Defaults to "euclidian".
            model_choice (str, optional): Is the final model you choose to embed your sequences. Defaults to 'Rostlab/prot_t5_xl_half_uniref50-enc'.
            show_strands (bool, optional): It means that you will plot a batch of the strands in your plot. Defaults to False.
            clone_size_factor (int, optional):  This value will be multiplied with the clone fractions of the sequences. If you set it to None, all nodes will have the same size. Defaults to 300.
            batch_size (int, optional): The number of sequences per samples which should be taken. The higher it is, the more computational intense.. Defaults to 1000.
            pca_components (int, optional): Number of principal components you reduce the 1024 dimensional output of the embedders to. Defaults to 50.
            prefered_cmap (str, optional): colormap for your points. Defaults to "viridis".
        """
        if samples == None:
            samples = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        UmapEmbed = protein_embedding_umap.PlotEmbedding(
            self.sequencing_report,
            samples,
            self.region_of_interest,
            show_strands,
            clone_size_factor,
            batch_size,
            pca_components,
            n_neighbors,
            min_dist,
            random_seed,
            densmap,
            metric,
            model_choice,
            ax=self.ControlFigure.ax,
            font_settings=self.font_settings,
            legend_settings=self.legend_settings,
            prefered_cmap=prefered_cmap,
            iterations_umap = iterations_umap
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test != True:
            self.save_cluster_report(
                cluster_report=UmapEmbed.umap_results, path=save_report_path
            )

    def umap_clustering_characteristic(
        self,
        samples=None,
        characteristic="length",
        n_neighbors=20,
        min_dist=0.01,
        random_seed=42,
        densmap=True,
        metric="euclidean",
        model_choice="Rostlab/prot_t5_xl_half_uniref50-enc",
        show_strands=False,
        clone_size_factor=300,
        batch_size=1000,
        pca_components=50,
        prefered_cmap="viridis",
        save_report_path=None,
        antigens=None,
        iterations_umap = 1000,
    ):
        """You can use this plot to cluster multiple samples based on certain characteristics such as sequence length hydrophobicity etc.

        Args:
            samples (str, optional): Here you can insert a list of samples you would like to compare and cluster. Defaults to None.
            characteristic (str, optional): Here you can define the characteristic you want to apply the supervised functionality from UMAP to (https://umap-learn.readthedocs.io/en/latest/supervised.html). If you take a look at the link y is the characteristic. You can insert aliphatic_index, isoelectric_point, hydrophobicity, weight, mass_charge_ratio, length or binding (will align the global data structure to binding values)
            n_neighbors (int, optional): Here you can choose the numbers of neighbors a point has.Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should be between 5 to 50. Defaults to 20.
            min_dist (float, optional): controls how tightly the points will be set to each other. It should be the minimum distance points are allowed to be apart from each other in the low dimensional representation. Defaults to 0.01.
            random_seed (int, optional): Set a certain seed for reprodubility. Defaults to 42.
            densmap (bool, optional): This parameter allows you to visualize points more densily which are also more dense in all dimensions to each other.. Defaults to True.
            metric (str, optional): You need to insert a string as input which is the distance metric for the UMAP algorithm. Defaults to "euclidian".
            model_choice (str, optional): Is the final model you choose to embed your sequences. Defaults to 'Rostlab/prot_t5_xl_half_uniref50-enc'.
            show_strands (bool, optional): It means that you will plot a batch of the strands in your plot. Defaults to False.
            clone_size_factor (int, optional):  This value will be multiplied with the clone fractions of the sequences. If you set it to None, all nodes will have the same size. Defaults to 300.
            batch_size (int, optional): The number of sequences per samples which should be taken. The higher it is, the more computational intense.. Defaults to 1000.
            pca_components (int, optional): Number of principal components you reduce the 1024 dimensional output of the embedders to. Defaults to 50.
            prefered_cmap (str, optional): colormap for your points. Defaults to "viridis".
        """

        if samples == None:
            samples = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        UmapEmbed = protein_embedding_umap.PlotEmbedding(
            self.sequencing_report,
            samples,
            self.region_of_interest,
            show_strands,
            clone_size_factor,
            batch_size=batch_size,
            pca_components=pca_components,
            n_neighbors=n_neighbors,
            min_dist=min_dist,
            random_seed=random_seed,
            densmap = densmap,
            metric = metric,
            model_choice = model_choice,
            characteristic=characteristic,
            ax=self.ControlFigure.ax,
            font_settings=self.font_settings,
            legend_settings=self.legend_settings,
            colorbar_settings=self.colorbar_settings,
            prefered_cmap=prefered_cmap,
            binding_data=self.binding_data,
            antigens=antigens,
            iterations_umap = iterations_umap
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test != True:
            self.save_cluster_report(
                cluster_report=UmapEmbed.umap_results, path=save_report_path
            )

    def cluster_binding_data_umap(
        self,
        samples=None,
        antigens=None,
        antigen_names=False,
        extra_figure=False,
        n_neighbors=20,
        min_dist=0.01,
        random_seed=42,
        densmap=True,
        metric="euclidean",
        model_choice="Rostlab/prot_t5_xl_half_uniref50-enc",
        show_strands=False,
        clone_size_factor=300,
        batch_size=1000,
        pca_components=50,
        prefered_cmap="viridis",
        save_report_path=None,
        iterations_umap = 1000,
    ):
        """You can use this plot to cluster your sequences with functional assay data based on UMAP

        Args:
            samples (str, optional): Here you can insert a list of samples you would like to compare and cluster. Defaults to None.
            antigens (str, optional): Names of the antigens with the corresponidng sequences you want to cluster with your ngs data. Input type is a list
            antigen_names (str, optional): Names of the antigens with the corresponidng sequences you want to cluster with your ngs data. Input type is a list
            characteristic (str, optional): Here you can define the characteristic you want to apply the supervised functionality from UMAP to (https://umap-learn.readthedocs.io/en/latest/supervised.html). If you take a look at the link y is the characteristic. You can insert aliphatic_index, isoelectric_point, hydrophobicity, weight, mass_charge_ratio, length or binding (will align the global data structure to binding values)
            n_neighbors (int, optional): Here you can choose the numbers of neighbors a point has.Larger values will result in more global structure being preserved at the loss of detailed local structure. In general this parameter should be between 5 to 50. Defaults to 20.
            min_dist (float, optional): controls how tightly the points will be set to each other. It should be the minimum distance points are allowed to be apart from each other in the low dimensional representation. Defaults to 0.01.
            random_seed (int, optional): Set a certain seed for reprodubility. Defaults to 42.
            densmap (bool, optional): This parameter allows you to visualize points more densily which are also more dense in all dimensions to each other.. Defaults to True.
            metric (str, optional): You need to insert a string as input which is the distance metric for the UMAP algorithm. Defaults to "euclidian".
            model_choice (str, optional): Is the final model you choose to embed your sequences. Defaults to 'Rostlab/prot_t5_xl_half_uniref50-enc'.
            show_strands (bool, optional): It means that you will plot a batch of the strands in your plot. Defaults to False.
            clone_size_factor (int, optional):  This value will be multiplied with the clone fractions of the sequences. If you set it to None, all nodes will have the same size. Defaults to 300.
            batch_size (int, optional): The number of sequences per samples which should be taken. The higher it is, the more computational intense.. Defaults to 1000.
            pca_components (int, optional): Number of principal components you reduce the 1024 dimensional output of the embedders to. Defaults to 50.
            prefered_cmap (str, optional): colormap for your points. Defaults to "viridis".
        """

        if samples == None:
            samples = [self.preferred_sample]
        if self.binding_data is None:
            print("Please upload binding data to use this functionality")
            return
        if antigens == None:
            antigens = [self.preferred_antigen]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        UmapEmbed = protein_embedding_umap.PlotEmbedding(
            self.sequencing_report,
            samples,
            self.region_of_interest,
            show_strands,
            clone_size_factor,
            batch_size,
            pca_components,
            n_neighbors,
            min_dist,
            random_seed,
            densmap,
            metric,
            model_choice,
            antigens=antigens,
            binding_data=self.binding_data,
            extra_figure=extra_figure,
            ax=self.ControlFigure.ax,
            font_settings=self.font_settings,
            legend_settings=self.legend_settings,
            prefered_cmap=prefered_cmap,
            colorbar_settings=self.colorbar_settings,
            toxin_names=antigen_names,
            iterations_umap = iterations_umap
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test != True:
            self.save_cluster_report(
                cluster_report=UmapEmbed.umap_results, path=save_report_path
            )

    def basic_cluster(
        self,
        samples=None,
        batch_size=1000,
        max_ld=1,
        min_ld=0,
        label_type="numbers",
        save_report_path=None,
    ):
        """
        :param samples: A list containing the samples you would like to analyze. Analyze just one sample with: ["My_sampel_name"].
        :param max_ld: Maximum allowed levenshtein distance between sequences within one cluster. The higher the distance the larger the clusters.
        :param min_ld: Minimum allowed levenshtein distance between sequences within one cluster.
        :param label_type: Default is numbers. This will label the nodes in the plot with the corresponding identifier in the output report. You can type sequences for labeling the nodes with the sequneces. If you do not want to have labels set it to None.
        :param save_report_path: Default is None which saves your report in my_experiments/reports_pipeline. If you want to change it somewhere else you need to insert the full path with filename.
        """
        if samples == None:
            samples = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        #  assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        cluster_report = levenshtein_clustering.LevenshteinClustering(
            self.sequencing_report,
            samples,
            self.ControlFigure.ax,
            self.region_of_interest,
            max_ld,
            min_ld,
            batch_size,
            label_type,
            self.font_settings,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test == True:
            self.save_cluster_report(
                cluster_report=cluster_report, path=save_report_path
            )

    def ls_distance_binding(
        self,
        samples=None,
        batch_size=1000,
        max_ld=1,
        min_ld=0,
        label_type="numbers",
        save_report_path=None,
        antigen_names=None,
    ):
        """
        :param samples: A list containing the samples you would like to analyze. Analyze just one sample with: ["My_sampel_name"].
        :param max_ld: Maximum allowed levenshtein distance between sequences within one cluster. The higher the distance the larger the clusters.
        :param min_ld: Minimum allowed levenshtein distance between sequences within one cluster.
        :param label_type: Default is numbers. This will label the nodes in the plot with the corresponding identifier in the output report. You can type sequences for labeling the nodes with the sequneces. If you do not want to have labels set it to None.
        :param save_report_path: Default is None which saves your report in my_experiments/reports_pipeline. If you want to change it somewhere else you need to insert the full path with filename.
        :param antigen_names: Names of the antigens with the corresponidng sequences you want to cluster with your ngs data. Input type is a list
        """
        if self.binding_data is None:
            print("Please upload binding data to use this functionality")
            return
        if samples == None:
            samples = [self.preferred_sample]
        if antigen_names == None:
            antigen_names = [self.preferred_antigen]

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        #  assert sample in self.experiments_list, "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        cluster_report = levenshtein_clustering.LevenshteinClustering(
            self.sequencing_report,
            samples,
            self.ControlFigure.ax,
            self.region_of_interest,
            max_ld,
            min_ld,
            batch_size,
            label_type,
            self.font_settings,
            self.binding_data,
            antigen_names,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test == True:
            self.save_cluster_report(
                cluster_report=cluster_report, path=save_report_path
            )

    def show(self):
        self.ControlFigure.fig.show()

    def cluster_binding_data(
        self,
        samples=None,
        strands=True,
        model="Rostlab/prot_t5_xl_half_uniref50-enc",
        pca_components=80,
        perplexity=30,
        iterations_tsne=2500,
        batch_size=1000,
        antigens=None,
        show_antigen_names=False,
        extra_figure=False,
        save_report_path=None,
    ):
        """
        :param samples: type is list. The samples you would like to compare towards their sequences
        :param strands: Default is True. It means that you will plot a batch of the strands in your plot
        :param pca_components: Default is 80. Has to be applied for better accuracy of t-SNE. You can indirectly change the described variance with this.
        :param model: The model you would like to choose for the embedding. Default is 'Rostlab/prot_t5_xl_half_uniref50-enc'. You can choose all models on https://huggingface.co/Rostlab and sgt.
        :param perplexity: Default is 30. It roughly determines the number of nearest neighbors that are considered in the embedding. A higher perplexity value results in a more global structure in the low-dimensional embedding, while a lower perplexity value emphasizes local structure. The optimal perplexity value for a given dataset depends on the dataset's intrinsic dimensionality, and it is usually determined by trial and err
        :param iterations_tsne: Default is 2500. number of times that the algorithm will repeat the optimization process for reducing the cost function. The optimization process aims to minimize the difference between the high-dimensional and low-dimensional representations of the data. More iterations result in a more optimized low-dimensional representation, but also increases the computational cost.
        :param batch_size: Default is 1000. The size of the sample which is chosen. The higher it is, the more computational intense.
        :param antigens: Datatype is a list. Is the names of the antigens in your binding data you would like to analyze. If you do not give any input, it will take the preferred antigen.
        :param show_antigen_names: Default is False. If you want to show the names of the antigens in the plot, set it to True.
        :param extra_figure: Default is False. If you want to have an extra figure with the indexes of the sequences in the vector space, set it to True.
        :return: Returns a plot where the sequences of the input samples are transformed in a vector space. Dimension reduction such as PCA and following t-SNE is used to plot it on a two dimensional space. The different colors indicate the different samples.
        """
        if samples == None:
            samples = [self.preferred_sample]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        assert (
            self.binding_data is not None
        ), "You have not given binding data. You can add it with the add_binding_data function"
        if antigens == None:
            antigens = [self.preferred_antigen]
        models_all = [
            "Rostlab/ProstT5_fp16",
            "Rostlab/prot_t5_xl_uniref50",
            "Rostlab/prot_t5_base_mt_uniref50",
            "Rostlab/prot_bert_bfd_membrane",
            "Rostlab/prot_t5_xxl_uniref50",
            "Rostlab/ProstT5",
            "Rostlab/prot_t5_xl_half_uniref50-enc",
            "Rostlab/prot_bert_bfd_ss3",
            "Rostlab/prot_bert_bfd_localization",
            "Rostlab/prot_electra_generator_bfd",
            "Rostlab/prot_t5_xl_bfd",
            "Rostlab/prot_bert",
            "Rostlab/prot_xlnet",
            "Rostlab/prot_bert_bfd",
            "Rostlab/prot_t5_xxl_bfd",
        ]
        assert (
            model in models_all
        ), f"Please enter a valid model name which are\n{models_all}. You can find nearly all of the models at: https://huggingface.co/Rostlab"
        if samples == None:
            samples = [self.experiments_list[0]]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert (
            not incorrect_samples
        ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        assert (
            type(samples) == list
        ), "You have to give a list with the samples you want to analyze"
        assert (
            type(strands) == bool
        ), "You have to give True or False as input for the strands parameter"
        assert (
            type(pca_components) == int
        ), "You have to give an integer as input for the pca_components"
        assert (
            type(perplexity) == int
        ), "You have to give an integer as input for the perplexity"
        assert (
            type(iterations_tsne) == int
        ), "You have to give an integer as input for the iterations_tsne"
        assert (
            type(batch_size) == int
        ), "You have to give an integer as input for the batch_size"
        assert (
            batch_size > pca_components
        ), "The batch_size has to be larger than the pca_components"
        assert (
            batch_size > perplexity
        ), "The batch_size has to be larger than the perplexity"
        assert (
            pca_components > perplexity
        ), "The pca_components has to be larger than the perplexity"
        assert (
            type(extra_figure) == bool
        ), "You have to give True or False as input for the extra_figure parameter"
        assert (
            type(antigens) == list
        ), "You have to give a list with the antigens you want to analyze"
        assert show_antigen_names in [
            True,
            False,
        ], "You have to give True or False as input for the show_antigen_names parameter"
        assert (
            type(save_report_path) == str or save_report_path == None
        ), "You have to give a string as input for the save_report_path parameter"

        self.ControlFigure.clear_fig()
        EmbeddingPlot = protein_embedding.PlotEmbedding(
            self.sequencing_report,
            model,
            samples,
            self.region_of_interest,
            strands,
            add_clone_size=300,
            batch_size=batch_size,
            pca_components=pca_components,
            perplexity=perplexity,
            iterations_tsne=iterations_tsne,
            antigens=antigens,
            font_settings=self.font_settings,
            legend_settings=self.legend_settings,
            binding_data=self.binding_data,
            colorbar_settings=self.colorbar_settings,
            ax=self.ControlFigure.ax,
            toxin_names=show_antigen_names,
            extra_figure=extra_figure,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

        if self.is_test == True:
            self.save_cluster_report(EmbeddingPlot.tsne_results, path=save_report_path)

    # @DeprecationWarning
    def cluster_one_AG(
        self,
        antigen=None,
        max_ld=1,
        min_ld=0,
        batch_size=1000,
        specific_experiments=False,
        prefered_cmap="Blues",
        label_type="numbers",
        save_report_path=None,
    ):
        warnings.warn(
            "This function will be removed in the future. Use the function cluster_binding_data instead."
        )
        """
        :param antigen: is the name of the antigen you would like to analyze in a list. You can also add multiple.
        :param max_ld: optional Parameter where its default is 1. Is the maximum Levenshtein distance you allow per cluster
        :param min_ld: optional Parameter where its default is 0. Is the minimum Levenshtein distance between sequences you allow
        :param batch_size: optional Parameter where its default is 1000. Is the batch size you want to use for the analysis
        :param specific_experiments: optional Parameter. You can give the names of specific samples in a list if you want
        :param preferred_cmap: optional Parameter. Define the colormap you would like to use.
        :param save_report_path: Default is None which saves your report in my_experiments/reports_pipeline. If you want to change it somewhere else you need to insert the full path with filename.
        :return: Creates a figure where sequences are clustered based on Levenshtein distance. Additionally the binding data of the sequences against a specific antigen is given.
        """

        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        assert (
            self.binding_data is not None
        ), "You have not given binding data. You can add it with the add_binding_data function"
        if antigen == None:
            antigen = self.preferred_antigen
        antigen = [antigen]
        assert (
            type(antigen) == list
        ), "You have to give a string as input for the antigen"
        assert (
            type(max_ld) == int
        ), "You have to give an integer as input for the maximum levenshtein distance"
        assert (
            type(min_ld) == int
        ), "You have to give an integer as input for the minimum levenshtein distance"
        if specific_experiments != False:
            assert (
                type(specific_experiments) == list
            ), "You have to give a list with the samples you want to analyze"
        else:
            specific_experiments = [self.preferred_sample]
        self.ControlFigure.clear_fig()
        cluster_report = levenshtein_clustering.LevenshteinClustering(
            self.sequencing_report,
            specific_experiments,
            self.ControlFigure.ax,
            self.region_of_interest,
            max_ld,
            min_ld,
            batch_size,
            label_type,
            self.font_settings,
            self.binding_data,
            antigen,
            prefered_cmap,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

        if self.is_test == True:
            self.save_cluster_report(
                cluster_report=cluster_report, path=save_report_path
            )

    # @DeprecationWarning
    def tsne_cluster_AG(
        self,
        sample=None,
        antigen=None,
        antigen_names=True,
        pca_components=70,
        perplexity=25,
        iterations_tsne=2500,
        save_report_path=None,
    ):
        """
        :param sample: the sample you would like to analyze
        :param antigen: Input is a list. the toxins you would like to cluster
        :param antigen_names: Default is True. Prints the name of the toxin for the corresponding embedded sequence in the plot
        :param pca_components: optional. Default is 70
        :param perplexity: optional. Default 25
        :param iterations_tsne: optional. Default is 2500
        :return: It first embeds the sequences in a vector space and then reduces the dimensions and clusters them with PCA and TSNE. The sequences with the binding data are processed with the input sequences, to enable the plotting of the binding data.
        """
        if sample == None:
            sample = self.preferred_sample
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        assert (
            self.binding_data is not None
        ), "You have not given binding data. You can add it with the add_binding_data function"
        if antigen == None:
            antigen = [self.preferred_antigen]
        assert type(sample) == str, "You have to give a string as input for the sample"
        assert (
            sample in self.experiments_list
        ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert antigen_names in [
            True,
            False,
        ], "You have to give True or False as input for the antigen_names parameter"
        assert (
            type(pca_components) == int
        ), "You have to give an integer as input for the pca_components"
        assert (
            type(perplexity) == int
        ), "You have to give an integer as input for the perplexity"
        assert (
            type(iterations_tsne) == int
        ), "You have to give an integer as input for the iterations_tsne"
        assert (
            type(antigen) == list
        ), "You have to give a list with the antigens you want to analyze"

        self.ControlFigure.clear_fig()
        tsne_results = embedding_with_binding.cluster_toxins_tsne(
            self.ControlFigure.fig,
            self.sequencing_report,
            sample,
            antigen,
            self.binding_data,
            antigen_names,
            pca_components,
            perplexity,
            iterations_tsne,
            self.font_settings,
            self.colorbar_settings,
            False,
            self.region_of_interest,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        if self.is_test == True:
            self.save_cluster_report(tsne_results, path=save_report_path)

    def embedding_network(
        self,
        samples=None,
        model="Rostlab/prot_t5_xl_half_uniref50-enc",
        batch_size=200,
        cmap="Blues",
        nodesize=300,
        threshold_distance=5,
    ):
        """
        :param samples: type is list. The samples you would like to compare towards their sequences
        :param model: The model you would like to choose for the embedding. You can choose "Rostlab/ProstT5_fp16",\n"Rostlab/prot_t5_xl_uniref50",\n"Rostlab/prot_t5_base_mt_uniref50",\n"Rostlab/prot_bert_bfd_membrane",\n"Rostlab/prot_t5_xxl_uniref50", \n"Rostlab/ProstT5", \n"Rostlab/prot_t5_xl_half_uniref50-enc", \n"Rostlab/prot_bert_bfd_ss3", \n"Rostlab/prot_bert_bfd_localization", \n"Rostlab/prot_electra_generator_bfd", \n"Rostlab/prot_t5_xl_bfd", \n"Rostlab/prot_bert", \n"Rostlab/prot_xlnet", \n"Rostlab/prot_bert_bfd", \n"Rostlab/prot_t5_xxl_bfd"
        :param batch_size: Default is 200. The size of the sample which is chosen. The higher it is, the more computational intense.
        :param nodesize: Default is 300. This value will be multiplied with the clone fractions of the sequences. If you set it to None, all nodes will have the same size.
        :threshold_distance: Default is 5. This value is multiplied with the summed batch_size of the samples. This defines the cluster size. The higher the values, the more complex the clusters.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert (
            type(threshold_distance) == int
        ), "Please enter an integer for the threshold distance"

        assert type(nodesize) == int, "Nodesize has to be an integer"
        assert nodesize > 0 & nodesize < 1000, "Please enter a value between 0 and 1000"
        protein_network_embedding.Network_Embedding(
            self.ControlFigure.ax,
            self.sequencing_report,
            samples,
            model,
            batch_size,
            self.font_settings,
            cmap,
            nodesize,
            threshold_distance,
            self.region_of_interest,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def embedding_tsne(
        self,
        samples=None,
        strands=True,
        model="Rostlab/prot_t5_xl_half_uniref50-enc",
        pca_components=80,
        perplexity=30,
        iterations_tsne=2500,
        batch_size=1000,
    ):
        """
        :param samples: type is list. The samples you would like to compare towards their sequences
        :param strands: Default is True. It means that you will visualize a batch of the strands in your plot
        :param pca_components: Default is 80. Has to be applied for better accuracy of t-SNE. You can indirectly change the described variance with this.
        :param model: The model you would like to choose for the embedding. Default is 'Rostlab/prot_t5_xl_half_uniref50-enc'. You can choose all models on https://huggingface.co/Rostlab.
        :param perplexity: Default is 30. It roughly determines the number of nearest neighbors that are considered in the embedding. A higher perplexity value results in a more global structure in the low-dimensional embedding, while a lower perplexity value emphasizes local structure. The optimal perplexity value for a given dataset depends on the dataset's intrinsic dimensionality, and it is usually determined by trial and err
        :param iterations_tsne: Default is 2500. number of times that the algorithm will repeat the optimization process for reducing the cost function. The optimization process aims to minimize the difference between the high-dimensional and low-dimensional representations of the data. More iterations result in a more optimized low-dimensional representation, but also increases the computational cost.
        :param batch_size: Default is 1000. The size of the sample which is chosen. The higher it is, the more computational intense.
        :return: Returns a plot where the sequences of the input samples are transformed in a vector space. Dimension reduction such as PCA and following t-SNE is used to plot it on a two dimensional space. The different colors indicate the different samples.
        """
        if samples == None:
            samples = [self.experiments_list[0]]
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        incorrect_samples = [x for x in samples if x not in self.experiments_list]
        assert (
            not incorrect_samples
        ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        assert (
            type(strands) == bool
        ), "You have to give True or False as input for the strands parameter"
        self.ControlFigure.clear_fig()
        protein_embedding.PlotEmbedding(
            self.sequencing_report,
            model,
            samples,
            self.region_of_interest,
            strands,
            add_clone_size=300,
            batch_size=batch_size,
            pca_components=pca_components,
            perplexity=perplexity,
            iterations_tsne=iterations_tsne,
            font_settings=self.font_settings,
            legend_settings=self.legend_settings,
            ax=self.ControlFigure.ax,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def morosita_horn(
        self,
        cmap="Blues",
        annotate_cells=False,
        specific_experiments=False,
        matrix_save_path=None,
    ):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_experiments: you can give a list with specific experiments
        :param matrix_save_path: Path where you want to save the matrix
        :return: Returns a matrix of the identity between your samples based on the Morosita Horn Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert (
                type(specific_experiments) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [
                x for x in specific_experiments if x not in self.experiments_list
            ]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        Matrix = make_matrix.IdentityMatrix(
            self.sequencing_report,
            self.region_of_interest,
            "morosita_horn",
            self.colorbar_settings,
            specific_experiments,
            self.ControlFigure.ax,
            True,
            self.font_settings,
            cmap,
            annotate_cells=annotate_cells,
        )

        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        save_matrix(Matrix.matrix, matrix_save_path)

    def jaccard(
        self,
        cmap="Blues",
        annotate_cells=False,
        specific_experiments=False,
        matrix_save_path=None,
    ):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_experiments: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the Jaccard Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert (
                type(specific_experiments) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [
                x for x in specific_experiments if x not in self.experiments_list
            ]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        Matrix = make_matrix.IdentityMatrix(
            self.sequencing_report,
            self.region_of_interest,
            "jaccard",
            self.colorbar_settings,
            specific_experiments,
            self.ControlFigure.ax,
            True,
            self.font_settings,
            cmap,
            annotate_cells=annotate_cells,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

        save_matrix(Matrix.matrix, matrix_save_path)

    def sample_diversity(self, method="InverseSimpson"):
        """
        :param method: Default is InverseSimpson. You can choose between InverseSimpson and Shannon which are two available indices to measure the diversity of the sequences in each of your samples based on the clone fraction.
        :return: Returns a barplot where the diversity of the sequences in each of your samples is shown.
        """
        assert method in [
            "InverseSimpson",
            "Shannon",
        ], "Please enter a valid method. You can choose between InverseSimpson and Shannon"
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        diversity_plot.DiversityPlot(
            self.sequencing_report,
            self.region_of_interest,
            self.ControlFigure.ax,
            self.font_settings,
            method=method,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def sorensen(
        self,
        cmap="Blues",
        annotate_cells=False,
        specific_experiments=False,
        matrix_save_path=None,
    ):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_samples: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the Sorensen Dice Index
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert (
                type(specific_experiments) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [
                x for x in specific_experiments if x not in self.experiments_list
            ]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        Matrix = make_matrix.IdentityMatrix(
            self.sequencing_report,
            self.region_of_interest,
            "sorensen",
            self.colorbar_settings,
            specific_experiments,
            self.ControlFigure.ax,
            True,
            self.font_settings,
            cmap,
            annotate_cells=annotate_cells,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        save_matrix(Matrix.matrix, matrix_save_path)

    def relative(self, cmap="Blues", annotate_cells=False, specific_experiments=False):
        """
        :param annotate_cells: Default is False. If you want to see the values of the matrix, set it to True.
        :param specific_samples: give a list with specific samples you would like to analyze
        :return: Returns a matrix of the identity between your samples based on the proportion of identical sequences
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        if specific_experiments != False:
            assert (
                type(specific_experiments) == list
            ), "You have to give a list with the samples you want to analyze"
            incorrect_samples = [
                x for x in specific_experiments if x not in self.experiments_list
            ]
            assert (
                not incorrect_samples
            ), f"The following sample(s) are not in your sequencing report: {', '.join(incorrect_samples)}. Please check the spelling or use the print_samples function to see the names of your samples"
        self.ControlFigure.clear_fig()
        Matrix = make_matrix.IdentityMatrix(
            self.sequencing_report,
            self.region_of_interest,
            "relative",
            self.colorbar_settings,
            specific_experiments,
            self.ControlFigure.ax,
            True,
            self.font_settings,
            cmap,
            annotate_cells=annotate_cells,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        save_matrix(Matrix.matrix)
        

    def levenshtein_dendrogram(self, samples=None, max_cluster_dist=2, batch_size=1000):
        """
        :params sample: the sample you would like to analyze in 
        :max_cluster_dist: Default is 2. Maximum levenshtein distance between sequences within a cluster.
        :params batch_size: Default is 1000. The size of the sample which is chosen.
        :return: Returns a dendrogram of the sequences based on the Levenshtein distance
        """
        if samples == None:
            samples = [self.preferred_sample]
        assert type(samples) == list, "You have to give a string as input for the sample"
        for sample in samples:
            assert (
                sample in self.experiments_list
            ), "The provided sample name is not in your sequencing report. Please check the spelling or use the print_samples function to see the names of your samples"
        assert (
            type(max_cluster_dist) == int
        ), "You have to give an integer as input for the maximum levenshtein distance"
        assert (
            type(batch_size) == int
        ), "You have to give an integer as input for the batch size"
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        hist_lvst_dist.LevenshteinDend(
            self.sequencing_report,
            self.region_of_interest,
            samples,
            self.font_settings,
            batch_size,
            max_cluster_dist,
            self.ControlFigure.ax,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def dendro_bind(
        self,
        sample=None,
        antigens=None,
        max_cluster_dist=2,
        batch_size=1000,
        ascending=True,
    ):
        """
        :params sample: the sample you would like to analyze
        :params antigens: the antigens you would like to analyze. The input is a list.
        :max_cluster_dist: Default is 2. Maximum levenshtein distance between sequences within a cluster. The higher this number the bigger the dendrogram
        :params batch_size: Default is 1000. The number of sequences starting with the highest fractions which are used for the analysis from the sample
        :params ascending: Default is True. If you want to see the highest fractions first, set it to False
        :return: Creates a dendrogram which shows the realtionship between your sequences in the sample and your sequences with binding data based on levenshtein distance. Additionally a barplot with the binding data of the found sequences is given.
        """
        if sample == None:
            sample = [self.preferred_sample]
        assert (
            self.binding_data is not None
        ), "You have not given binding data. You can add it with the add_binding_data function"
        if antigens == None:
            antigens = [self.preferred_antigen]


        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "multi"
        self.ControlFigure.clear_fig()
        fig2 = hist_lvst_dist_bind.DendroBind(
            self.sequencing_report,
            sample,
            self.region_of_interest,
            antigens,
            batch_size,
            max_cluster_dist,
            self.binding_data,
            ascending,
            self.ControlFigure.fig,
            self.font_settings,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )
        #  self.ControlFigure.tighten()
        return fig2

    def connect_samples(
        self,
        summed_clonefraction=0.5,
        max_num_reads=100,
        color_lines="peachpuff",
        max_weight_lines=100,
    ):
        """
        :param summed_clonefraction: You will take the top 50% of clones per sample per default. You can change that to take more or less crones
        :param max_num_reads: You will limit the maximum number of clones to 100 per default. If you want to have more your can increase that
        :param max_weight_lines: If you have a levenshtein distance of 0 between two sequences this value will be the thickness of the line. The thickness of other distances is indicated through: max_weight_lines * (1 / ls_distance ** 2)
        :result: A network plot which shows the relation between the samples for the chosen clones based on levenshtein distance. The thickness of the lines indicates the inverse levenshtein distance.
        """
        self.ControlFigure.check_fig()
        self.ControlFigure.plot_type = "single"
        self.ControlFigure.clear_fig()
        sample_cluster.ClusterExperiment(
            self.sequencing_report,
            self.region_of_interest,
            summed_clonefraction=summed_clonefraction,
            max_num_reads=max_num_reads,
            ax=self.ControlFigure.ax,
            edge_color=color_lines,
            max_weight_lines=max_weight_lines,
        )
        self.ControlFigure.update_plot()
        self.style = plot_styler.PlotStyle(
            self.ControlFigure.ax, self.ControlFigure.plot_type
        )

    def validate_mixcr_path(self):
        path_to_mixcr = self.global_params["mixcr_path"]
        settings_dir = os.path.join(self.module_dir, "settings", "global_vars.txt")
        check_mixcr(path_to_mixcr, self.global_params, settings_dir)

    def create_parser(self):
        self.validate_mixcr_path()
        path_to_mixcr = self.global_params["mixcr_path"]
        commands = []
        commands.extend(
            ["java", f"-Xms{self.java_heap_size}M", "-jar"]
        )  # enable change of para
        commands.extend([path_to_mixcr])
        return commands

    def export_mixcr_quality(
        self, save_dir=None, specific_chain=None, metric=None, plot_type=None
    ):
        if save_dir == None:
            save_dir = self.mixcr_plots_path
        output_file = os.path.join(save_dir, self.experiment + "_alignQc.pdf")
        path_to_tables = os.path.join(
            self.module_dir,
            "my_experiments",
            self.experiment,
            "clones_result",
            "*.clns",
        )
        export_plots_commands = self.create_parser()

        export_plots_commands.extend(["exportQc"])
        export_plots_commands.extend(["align"])
        export_plots_commands.extend([path_to_tables])
        export_plots_commands.extend([output_file])
        export_plots_commands.extend(["--force-overwrite"])
        subprocess.run(export_plots_commands)
        print(f"You can find the result file at: {output_file}")

        #    for i in glob.glob(path_to_tables):

        #       name=os.path.basename(i).split(".")[0]
        #      output_file = os.path.join(save_dir, name + "_tags.pdf")
        #     export_plots_commands = self.create_parser()
        #    export_plots_commands.extend(["exportQc"])
        #   export_plots_commands.extend(["tags"])
        #   export_plots_commands.extend([i])
        #   export_plots_commands.extend([output_file])
        #   export_plots_commands.extend(["--force-overwrite"])
        #  subprocess.run(export_plots_commands)
        # print(f"You can find the result file at: {output_file}")
        #    name = os.path.basename(i).split(".")[0]
        #   output_file = os.path.join(save_dir, name + "chainUsage.png")
        #  export_plots_commands = self.create_parser()
        #  export_plots_commands.extend(["exportQc"])
        # export_plots_commands.extend(["chainUsage"])
        # export_plots_commands.extend(["--hide-non-functional"])
        # export_plots_commands.extend([i])
        # export_plots_commands.extend([output_file])
        # export_plots_commands.extend(["--force-overwrite"])
        # subprocess.run(export_plots_commands)
        # print(f"You can find the result file at: {output_file}")

        output_file = os.path.join(save_dir, self.experiment + "_coverage.png")
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportQc"])
        export_plots_commands.extend(["coverage"])
        export_plots_commands.extend([path_to_tables])
        export_plots_commands.extend([output_file])
        export_plots_commands.extend(["--force-overwrite"])
        subprocess.run(export_plots_commands)
        print(f"You can find the result file at: {output_file}")

        if specific_chain != None:
            export_plots_commands.extend(["--chains", specific_chain])
        if metric != None:
            export_plots_commands.extend(["--metrics", metric])
        if plot_type != None:
            export_plots_commands.extend(["--plotType", plot_type])

    def mixcr_explain_diversity(
        self,
        plot_name="cdr3_metrics",
        plot_save_dir=None,
        metric="cdr3metrics",
        chains=None,
        plot_type="boxplot",
        output_type="pdf",
    ):
        """
        :param: metric: Default is cdr3metrics. You can choose between cdr3metrics and diversity. If you choose diversity you will focus on the clone fractions whereas with cdr3metrics you will focus on certain aspects of these chains such as length (https://mixcr.com/mixcr/reference/mixcr-postanalysis/#diversity-measures)
        :param: chains: Default is None. Possible inputs are : IGH, IGL, IGK, TRA, TRB, TRG, TRD, IG. You will choose a specific chain for the analysis.
        :param: plot_type: Default is violin. You can choose between: boxplot, boxplot-bindot, boxplot-jitter, violin, violin-bindot, barplot, barplot-stacked, lineplot, lineplot-jitter, lineplot-bindot, scatter
        :param: output_type: Default is pdf. Allows you to control the file type of your output. Possible values are: pdf, jpeg, png, svg, eps,
        """
        if metric in ["cdr3metrics", "diversity"]:
            pass
        else:
            print(
                "Please choose a correct value for the metric. Pipeline continues with:\ncdr3metrics and cdr3lenAA for metric_type"
            )
            metric = "cdr3metrics"
            metric_type = "cdr3lenAA"

        chains_values = [None, "IGH", "IGL", "IGK", "TRA", "TRB", "TRG", "TRD", "IG"]
        if chains not in chains_values:
            print(
                "Please enter a correct shortcut for the chain preferred chain. These are: "
            )
            [print(i) for i in chains_values]
        plot_type_values = [
            "boxplot",
            "boxplot-bindot",
            "boxplot-jitter",
            "violin",
            "violin-bindot",
            "barplot",
            "barplot-stacked",
            "lineplot",
            "lineplot-jitter",
            "lineplot-bindot",
            "scatter",
        ]
        if plot_type not in plot_type_values:
            print(
                "Please enter a correct shortcut for the chain plot type. These are: "
            )
            [print(i) for i in plot_type_values]

        print(
            f"To be able to understand the following processing go to: https://mixcr.com/mixcr/reference/mixcr-postanalysis/#overlap-postanalysis"
        )
        plot_name = plot_name

        if plot_save_dir == None:
            plot_save_dir = self.mixcr_plots_path
        save_dir = os.path.join(self.module_dir, "my_experiments", self.experiment)
        json_dir = os.path.join(save_dir, "pa_individual.json")
        path_to_tables = os.path.join(
            self.module_dir,
            "my_experiments",
            self.experiment,
            "clones_result",
            "*.clns",
        )
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["individual"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend([metric])
        #  export_plots_commands.extend(["--metric",metric_type])
        #  export_plots_commands.extend(["-f"])
        if not chains == None:
            export_plots_commands.extend(["--chains", chains])
        export_plots_commands.extend(["--plot-type", plot_type])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend(
            [os.path.join(plot_save_dir, plot_name + "." + output_type)]
        )
        subprocess.run(export_plots_commands)

    def mixcr_vdj_usage(
        self,
    ):
        save_dir = self.mixcr_plots_path

        json_dir = os.path.join(
            self.module_dir, "my_experiments", self.experiment, "pa_individual.json"
        )
        path_to_tables = os.path.join(
            self.module_dir,
            "my_experiments",
            self.experiment,
            "clones_result",
            "*.clns",
        )
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["individual"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)

        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["vUsage"])
        #  export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "vUsage_heatmap.png")])
        subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["jUsage"])
        # export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend([os.path.join(save_dir, "jUsage_heatmap.png")])
        subprocess.run(export_plots_commands)
        export_plots_commands = self.create_parser()
        export_plots_commands.extend(["exportPlots"])
        export_plots_commands.extend(["isotypeUsage"])
        # export_plots_commands.extend(["-f"])
        export_plots_commands.extend([json_dir])
        export_plots_commands.extend(
            [os.path.join(save_dir, "isotypeUsage_heatmap.png")]
        )
        subprocess.run(export_plots_commands)

    def mixcr_overlap(self):
        save_dir = self.plot_path
        path_to_tables = os.path.join(
            self.module_dir, "my_experiments", self.experiment, "clones_result"
        )
        json_dir = os.path.join(
            self.module_dir, "my_experiments", self.experiment, "pa_overlap.json"
        )
        if not os.path.isfile(json_dir):
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["postanalysis"])
            export_plots_commands.extend(["overlap"])
            export_plots_commands.extend(["--default-downsampling", "none"])
            export_plots_commands.extend(["--default-weight-function", "none"])
            export_plots_commands.extend([path_to_tables])
            export_plots_commands.extend([json_dir])
            subprocess.run(export_plots_commands)
        possible_metrics = [
            "SharedClonotypes",
            "RelativeDiversity",
            "F1Index",
            "F2Index",
            "JaccardIndex",
            "Pearson",
            "PearsonAll",
        ]

        for metric in possible_metrics:
            export_plots_commands = self.create_parser()
            export_plots_commands.extend(["exportPlots"])
            export_plots_commands.extend(["overlap"])
            export_plots_commands.extend(["--metric", metric])
            export_plots_commands.extend(["--color-key", "Patient"])
            export_plots_commands.extend([json_dir])
            export_plots_commands.extend([os.path.join(save_dir, f"{metric}.png")])
            subprocess.run(export_plots_commands)

    # export_plots_commands = self.create_parser()
    # export_plots_commands.extend(["exportPlots"])
    # export_plots_commands.extend(["overlap"])


#     export_plots_commands.extend(["--metric", metric])
# export_plots_commands.extend(["--color-key", "Patient"])
# export_plots_commands.extend([json_dir])
# export_plots_commands.extend([os.path.join(save_dir, f"overlap.png")])
# subprocess.run(export_plots_commands)


def automation(make_report=False):
    plot = PlotManager()
    regions = plot.avail_regions
    if "targetSequences" in regions:
        regions.remove("targetSequences")
    for region in regions:
        plot.change_region(region=region)
        plot.full_analysis()
        if make_report == True:
            plot.create_report()



from .modelling.nanonet import Use_NanoNet
from .modelling.create_fasta import Fasta
from modelling import multi_sequence_align
from .settings.change_settings import Settings
from tkinter.filedialog import askopenfilename
import subprocess
import sys

class ModelProtein:
    
    def __init__(self) -> None:
        Sets = Settings.check_dirs()
        if self.git_available() == False:
            raise Exception("Please install git to be able to use this module.")
        
    def model_nanonet(modeller = False):
        """
        :param modeller: If set to true modeller will be used for side chain reconstruction.
        :result: Output is a pdb file which is the predicted 3D structure from the linear sequence which is the input.
        """
        fasta_path = askopenfilename()
        is_fasta = Fasta.check_fasta(fasta_path)
        if is_fasta:
            Use_NanoNet(fasta_path, modeller)
        
    def create_fasta(sequence, sequence_name, fasta_name):
        Fasta.make_fasta(sequence, sequence_name, fasta_name)
        
    def multi_sequence_alignment(fasta_path, out_path):
        multi_sequence_align.MSA(fasta_path, out_path)
        
    
    @staticmethod
    def git_available():
        try:
            output = subprocess.check_output(['git', '--version'], stderr=subprocess.STDOUT, text=True)
            print(f"Git is available. Version: {output.strip()}")
            git_avail = True
        except subprocess.CalledProcessError as e:
            print("Git is not available.")
            git_avail = False
            return git_avail
        


import subprocess
from modeller import *
import os
from .create_fasta import Fasta



# LIcence states: "Research Use" means research, evaluation, or development for the purpose of advancing knowledge, teaching, learning, or customizing the Technology or Modifications for personal use. Research Use expressly excludes use or distribution for direct or indirect commercial (including strategic) gain or advantage.
class Use_NanoNet:
    def __init__(self,module_dir, fasta_path, modeller = False) -> None:
        self.module_dir = module_dir
        self.git_available()
       # self.check_tensorflow()
        self.prepare_NanoNet(fasta_path)
        Fasta_gen = Fasta()
        if Fasta_gen.check_fasta(fasta_path):
            avail_modeller = self.add_modeller()
            if modeller and avail_modeller:
                command_modeller = ["-m"]
                self.base_commands += command_modeller
            if os.path.isfile(os.path.join(self.module_dir, "NanoNet","NanoNet.py")):
                print("Please cite:")
                self.print_citation()
                subprocess.run(self.base_commands)
            else:
                print("Cannot access NanoNet repo.")
        else:
            print("No valid fasta file.")
    def git_available(self):
        try:
            output = subprocess.check_output(['git', '--version'], stderr=subprocess.STDOUT, text=True)
            print(f"Git is available. Version: {output.strip()}")
            self.clone_repo()
        except subprocess.CalledProcessError as e:
            print("Git is not available.")
            return 
    
    @staticmethod
    def print_citation():
        dir = os.path.dirname(__file__)
        with open(os.path.join(dir, "citations", "nanonet.txt")) as f:
            text = f.read()

            # Print the text
            print(text)

            # Close the file
            f.close()
    
    @staticmethod
    def check_tensorflow():
        try:
            import tensorflow as tf
            print(f"TensorFlow version: {tf.__version__}")
        except ImportError:
            print("TensorFlow is not installed. Run pip install tensorflow in your terminal")      
            subprocess.run(["pip install tensorflow"])

    def clone_repo(self):
        if not os.path.isdir(os.path.join(self.module_dir, "NanoNet")):
            subprocess.run(["git", "clone", "https://github.com/dina-lab3D/NanoNet.git"])
        else:
            pass
    def prepare_NanoNet(self, fasta_file):
        self.base_commands = ["python", 
                        os.path.join(self.module_dir, "NanoNet","NanoNet.py"),
                        "-t",
                        os.path.abspath(fasta_file),
                        ]
        
    def add_modeller(self, ):
        try:
            env = environ()
            aln = alignment(env)
            avail_modeller = True
        except NameError:
            print("Modeller is not installed. You can continue without.")
            avail_modeller = False
        return avail_modeller
            
    

Use_NanoNet(module_dir = r"C:\Users\nilsh\my_projects\ExpoSeq", fasta_path=r"C:\Users\nilsh\my_projects\ExpoSeq\NanoNet\small_test.fa", modeller = True )


import matplotlib.pyplot as plt
import networkx as nx
import math
import pandas as pd
import community.community_louvain as community
import editdistance


class PrepareData:
    @staticmethod
    def check_input(samples, max_ld, min_ld, batch_size, binding_data, antigens):
        assert (
            type(samples) == list
        ), "You have to give a string as input for the sample"
        assert (
            type(max_ld) == int
        ), "You have to give an integer as input for the maximum levenshtein distance"
        assert (
            type(min_ld) == int
        ), "You have to give an integer as input for the minimum levenshtein distance"
        assert type(batch_size) == int
        assert batch_size > 1
        assert max_ld > min_ld
        assert max_ld >= 1
        if binding_data is not None:
            assert type(antigens) == list
            for antigen in antigens:
                assert antigen in binding_data.columns.tolist()

    @staticmethod
    def clean_data(
        sequencing_report,
        samples,
        batch_size,
        experiment_column,
        binding_data,
        antigens,
        region_of_interest,
    ):
        sample_report = sequencing_report.loc[
            sequencing_report[experiment_column].isin(samples)
        ]
        sample_report = sample_report.groupby(experiment_column).head(batch_size)
        sample_report = sample_report.drop_duplicates(subset=[region_of_interest])
        if binding_data is not None:
            merged_columns = [region_of_interest] + antigens
            binding_data = binding_data[merged_columns]
            mix = sample_report.merge(binding_data, on=region_of_interest, how="outer")
            sample_report = mix.fillna(0)

        return sample_report

    def calc_edges(
        self,
        sequencing_report,
        samples,
        batch_size=1000,
        max_ld=2,
        min_ld=0,
        region_string="aaSeqCDR3",
        binding_data=None,
        antigens=None,
        experiment_column="Experiment",
    ):
        """Create the Graph object for networkx which can be used to visualize the clusters

        Args:
            sequencing_report (pd.DataFrame):  report which contains all sequencing information from all samples after the processing with mixcr
            samples (list): List containing the names of the samples which should be used for the final plot.
            batch_size (_type_): Equals to the number of sequences which are drawn from the sequencing report for the embedding and the final plot. Defaults to 1000.
            max_ld (int): Is the maximum allowed difference between two sequences based on levenshtein distance to be still in one cluster. Default is 2.
            min_ld (int): Is the minimum allowed levenshtein distance between two sequence to be still in one cluster. Default is 0.
            region_string (str): A string which indicates the column name from which the amino acid sequences should be taken from.
            binding_data (pd.DataFrame, optional): Dataframe which contains the sequences and the binding values to the antigens. Defaults to None.
            antigens (list, optional): list containing the names of the antigens which are the headers of the binding data. From these antigens the binding data will be taken for the plot. Defaults to None.
            experiment_column (str, optional): Name of the column which contains the sample names in the sequencing report. Defaults to "Experiment".

        Returns:
            G: Graph object for networkx containing the nodes and edges.
            report: tidied report as input for the threshold algorithm based on levenshtein distance
        """
        self.check_input(samples, max_ld, min_ld, batch_size, binding_data, antigens)
        report = self.clean_data(
            sequencing_report,
            samples,
            batch_size,
            experiment_column,
            binding_data,
            antigens,
            region_string,
        )
        G = nx.Graph()
        aa = list(report[region_string])
        num_strings = len(aa)
        string_index = [*range(0, num_strings, 1)]
        max_ld = max_ld
        min_ld = min_ld
        for seq in aa:
            G.add_node(seq)
        for i in range(len(aa)):
            string1 = aa[i]
            index_1 = i
            for index in range(index_1):
                string2 = aa[index]
                distance = editdistance.distance(string1, string2)
                if (
                    distance >= min_ld and distance <= max_ld
                ):  # sequences are not added to Graph if tehy dont meet the threshold requirements
                    G.add_edge(string1, string2)
        G_deg = G.degree()
        to_remove = [
            n for (n, deg) in G_deg if deg == 0
        ]  # degree = 0 means that these nodes are not connected to any other nodes. Since they have their own cluster they can be removed
        G.remove_nodes_from(to_remove)
        sequences_clustered = list(G.nodes)
        report = report[
            report[region_string].isin(sequences_clustered)
        ]  # remove those  which are not in self.G
        return G, report


class LevenshteinClustering:
    def __init__(
        self,
        sequencing_report,
        samples,
        ax=None,
        region_string="aaSeqCDR3",
        max_ld=2,
        min_ld=0,
        batch_size=200,
        label_type="numbers",
        font_settings={},
        binding_data=None,
        antigens=None,
        prefered_cmap="viridis",
    ):
        self.ax = ax
        self.G, report = PrepareData().calc_edges(
            sequencing_report,
            samples,
            batch_size,
            max_ld,
            min_ld,
            region_string,
            binding_data,
            antigens,
        )
        self.report = report
        nodesize, label_numbers, label_sequences = self.calculate_nodesize(
            report, region_string, self.G
        )
        label_sequences = self.create_sequence_labels(self.G)
        partition = self.get_partition(self.G)
        if ax != None:
            partition = self.create_network(
                label_type,
                label_numbers,
                label_sequences,
                nodesize,
                partition,
                region_string,
                report,
                binding_data,
                antigens,
                prefered_cmap,
            )
            if font_settings != {}:
                self.add_header(font_settings, samples)

        self.cluster_report = self.generate_report(partition)
        self.nodesize = nodesize
        self.label_sequences = label_sequences
        self.partition = partition
    @staticmethod
    def calculate_nodesize(
        sample_report, region_string, G, cf_column_name="cloneFraction"
    ):
        """Calculates the nodesize and sets the labels for the corresponding sequences. The label can be either a number which indicates the id of the sequence in the report (sample_report) or a sequence.

        Args:
            sample_report (pd.DataFrame): is the report as output from PrepareData.tidy()
            region_string (str): A string which indicates the column name from which the amino acid sequences should be taken from.
            G (networkx.Graph): Graph object from networkx which contains the nodes and the edges for the plot.
            cf_column_name (str, optional):  Name of the column which contains the clone fraction in the sequencing report. Defaults to "cloneFraction".

        Returns:
            nodesize (list): size of the nodes in order
            label_numbers (dict): dictionary containing the ids for the corresponding sequences
            label_sequencse (dict): dicitinoary containing the sequences as label for the corresponding sequences
        """
        nodesize = []
        label_numbers = {}
        label_sequences = {}
        for index, g in enumerate(G):
            for i in sample_report[region_string]:
                if i == g:
                    node = math.sqrt(
                        sample_report.loc[
                            sample_report[region_string] == i, cf_column_name
                        ].values[0]
                    )
                    if node == 0:
                        node = 0.5  # this is for the once which do not have a clone count, so the binding data. otherwise they are not visible
                    nodesize.append(node)

                    if node * 500 > 1 / 100 * 500:
                        label_numbers[g] = index
                        label_sequences[g] = g
                    else:
                        label_numbers[g] = ""
                        label_sequences[g] = ""
                    break
        for y, n in enumerate(nodesize):
            nodesize[y] = int(n * 500)

        return nodesize, label_numbers, label_sequences

    @staticmethod
    def create_sequence_labels(G):
        ## labels
        label_sequences = {}
        node_ids = list(G.nodes())
        n = 0
        for index, g in enumerate(G):
            if n == 9:
                label_sequences[g] = g
                n = 0
            else:
                label_sequences[g] = ""
            n += 1
        return label_sequences

    @staticmethod
    def get_partition(G):
        """This function is used to partition the graph G into communities using the Louvain method, which is a popular algorithm for community detection. It returns a dictionary where the keys are the nodes of the graph G, and the values are the community assignments for each node.

        Args:
            G (networkx.Graph): Graph object from networkx which contains the nodes and the edges for the plot.

        Returns:
            _type_: _description_
        """
        partition = community.best_partition(G)
        return partition

    @staticmethod
    def map_binding(
        G, binding_data, region_of_interest, report, antigens, prefered_cmap, partition
    ):
        """_summary_

        Args:
            G (networkx.Graph): Graph object from networkx which contains the nodes and the edges for the plot.
            binding_data (pd.DataFrame, optional): Dataframe which contains the sequences and the binding values to the antigens. Defaults to None.
            region_of_interest (str): A string which indicates the column name from which the amino acid sequences should be taken from.
            report: tidied report from PrepareData().tidy()
            antigens (list, optional): list containing the names of the antigens which are the headers of the binding data. From these antigens the binding data will be taken for the plot. Defaults to None.
            prefered_cmap (str): cmap for the colorbar
            partition (dict): output from LevenshteinClustering.get_partition()
        """
        node_colors = {}
        assert (
            region_of_interest in binding_data.columns
        ), "You do not have binding data for the corresponding region of interest"
        # new_df = binding_data.loc[:, antigens]
        new_df = report.loc[:, antigens]
        maximum_binding_value = new_df.max().max()
        minimum_binding_value = new_df.min().min()
        if maximum_binding_value == 0:
            print("no binding values mapped")
            node_colors_norm_heatmap = list(partition.values())
            sm = None
        else:
            cmap = plt.colormaps.get_cmap(prefered_cmap)
            norm = plt.Normalize(minimum_binding_value, maximum_binding_value)
            sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)  # for colorbar
            for i, g in enumerate(G):
                sequence_row = report[report[region_of_interest] == g]
                right_binding_value = max(
                    [sequence_row[column].iloc[0] for column in antigens]
                )  # if you have multiple antigens this will find the right value to map
                sequence = sequence_row[region_of_interest].iloc[0]
                node_colors[sequence] = right_binding_value
            node_colors_norm_heatmap = [
                cmap(norm(node_colors[node])) if node_colors[node] != 0 else "gray"
                for node in G.nodes()
            ]  # all nodes without binding values are gray
        return node_colors_norm_heatmap, sm, maximum_binding_value

    def create_network(
        self,
        label_type,
        label_numbers,
        label_sequences,
        nodesize,
        partition,
        region_of_interest,
        report,
        binding_data,
        antigens,
        prefered_cmap,
    ):
        if binding_data is not None:
            node_colors, sm, maximum_binding_value = self.map_binding(
                self.G,
                binding_data,
                region_of_interest,
                report,
                antigens,
                prefered_cmap,
                partition,
            )
        else:
            maximum_binding_value = 0
            node_colors = list(partition.values())

        pos = nx.spring_layout(self.G)
        nx.draw_networkx_nodes(
            self.G,
            pos,
            node_size=nodesize,
            cmap=plt.cm.RdYlBu,
            node_color=node_colors,
            ax=self.ax,
        )
        nx.draw_networkx_edges(self.G, pos, alpha=0.3, ax=self.ax)
        if label_type == "numbers":
            label = label_numbers
        else:
            label = label_sequences

        nx.draw_networkx_labels(
            self.G,
            pos,
            labels=label,
            alpha=0.5,
            font_color="grey",
            horizontalalignment="right",
            verticalalignment="bottom",
            ax=self.ax,
        )
        if maximum_binding_value > 0:
            plt.colorbar(sm, ax=self.ax)
        return partition

    def add_header(self, font_settings, samples):
        original_fontsize = font_settings["fontsize"]
        font_settings["fontsize"] = 22
        self.ax.set_title(
            f"Connected components of {', '.join(samples)}", **font_settings
        )
        font_settings["fontsize"] = original_fontsize
        self.ax.set_axis_off()

    @staticmethod
    def generate_report(partition):
        cluster_report = pd.DataFrame([partition]).T
        cluster_report["Sequences"] = cluster_report.index
        cluster_report.reset_index(drop=True, inplace=True)
        cluster_report.rename(columns={0: "Cluster No."}, inplace=True)
        return cluster_report


# import pandas as pd
# file = r"C:\Users\nilsh\OneDrive\Desktop\DTU\NGS_pipeline\data\Binding_data\Chris_main_df.csv"
# binding_data = pd.read_csv(file)
# antigens = ["Ecarpholin_S_DDEL_5ug/mL", "Myotoxin_II_DDEL_5ug/mL"]
# filename = r"C:\Users\nilsh\my_projects\ExpoSeq\my_experiments\test_all_samples\sequencing_report.csv"
# sequencing_report = pd.read_csv(filename)
# sequencing_report["cloneFraction"] = sequencing_report["readFraction"]
# fig = plt.figure(1)
# ax = fig.gca()
# samples = ["Library_1_F11_2", "test_1_F10"]
# font_settings = {'fontfamily': 'serif', 'fontsize': '18', 'fontstyle': 'normal', 'fontweight': 'bold'}
# LevenshteinClustering(sequencing_report, samples, ax, font_settings=font_settings, batch_size=800, binding_data=binding_data, antigens = antigens)
# plt.show()



import pandas as pd
from .tidy_cluster_embedding import tidy_embed
from textwrap import wrap
import pandas as pd
from sgt.SGT import SGT
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

class ClusterEmbedding:
    def __init__(self) -> None:
        pass
    



    def tidy_embed(sequencing_report, batch_size, list_experiments, pca_components, perplexity ,iterations_tsne, region_string):
        report_batch = sequencing_report.groupby("Experiment").head(batch_size)
        selected_rows = report_batch.loc[report_batch["Experiment"].isin(list_experiments)]
        sgt = SGT(kappa=1,
                lengthsensitive=True)
        sequences = selected_rows[region_string].map(list)
        sequences_list = list(sequences)
        sgtembedding_df = sgt.fit_transform(corpus=sequences_list)
        pca = PCA(n_components=pca_components)
        pca.fit(sgtembedding_df)
        X = pca.transform(sgtembedding_df)
        print(np.sum(pca.explained_variance_ratio_))
        tsne = TSNE(n_components=2,
                    verbose=0,
                    perplexity=perplexity,
                    n_iter=iterations_tsne)
        tsne_results = tsne.fit_transform(X)
        tsne_results = pd.DataFrame(tsne_results,
                                    columns=[["tsne1", "tsne2"]])
        aminoacids = selected_rows["aaSeqCDR3"].to_list()
        experiments_batch = selected_rows.Experiment
        unique_experiments_num = pd.factorize(experiments_batch)[0]
        tsne_results["experiments_factorized"] = list(unique_experiments_num)
        tsne_results["experiments_string"] = list(experiments_batch)
        return tsne_results, aminoacids, experiments_batch
    
    
    

def show_difference(sequencing_report, list_experiments,strands, batch_size, pca_components, perplexity, iterations_tsne,region_string, ax, legend_settings, font_settings):
    tsne_results, aminoacids, experiments_batch = tidy_embed(sequencing_report,
                                                              batch_size,
                                                              list_experiments,
                                                              pca_components,
                                                              perplexity,
                                                              iterations_tsne,
                                                              region_string)
    tsne_plot = ax.scatter(tsne_results.tsne1,
                           tsne_results.tsne2,
                           c = pd.factorize(experiments_batch)[0],
                           alpha = 0.5
                            )
    ax.legend(handles=tsne_plot.legend_elements()[0],
               labels=list_experiments,
              **legend_settings)
    x = tsne_results["tsne1"].values.tolist()
    y = tsne_results["tsne2"].values.tolist()
    ax.set_xlabel("t-SNE1", **font_settings)
    ax.set_ylabel("t-SNE2", **font_settings)
    if strands == True:
        for i in range(0, len(x), 10):
            ax.annotate(aminoacids[i],
                        (x[i][0], y[i][0]),
                        fontsize = 5,
                        )
    original_fontsize = font_settings["fontsize"]
    font_settings["fontsize"] = 22
    title = "\n".join(wrap("t-SNE embedding of for given samples", 40))
    ax.set_title(title, pad= 12, **font_settings)
    font_settings["fontsize"] = original_fontsize


import pandas as pd




class FullSequence:
    def __init__(self, avail_regions):
        self.region_order = ["FR1", "CDR1", "FR2", "CDR2", "FR3", "CDR3", "FR4"]

        self.avail_regions = avail_regions
        
        
    
    @staticmethod
    def find_most_ones(matrix_df):
        column_sums = matrix_df.sum(axis=0)  # Sum along columns
        most_connected_column = column_sums.idxmax()  # Index of the column with the highest sum
        n = 0
        z = 0
        col_list = matrix_df[most_connected_column].to_list()
        col_series = matrix_df[most_connected_column]
        for i in range( len(col_list)):
            value = col_list[i]
            if value ==1:
                n += 1
            else:
                n = 0
            fragment = matrix_df.index[i]
            if n > z or ( fragment == "CDR3" and n == z):
                end_ind = i 
                z = n
        end_ind += 1
        start = end_ind - z 
        assert start >=0
        row_indexes = col_series.index[start:end_ind].to_list()
        return row_indexes
    
    def find_connecting_seq(self):
        region_order = ["FR1", "CDR1", "FR2", "CDR2", "FR3", "CDR3", "FR4"]
        region_matrix = pd.DataFrame(0, index=region_order, columns=region_order)
        len_matrix = len(region_order)
        for region_index in range(len_matrix): # there is a more elegant way if you use the avail regions as array but this was quicker
            consecutive = 0
            for region_row in range(len_matrix):
                index_matrix = region_matrix.index[region_row]
                if index_matrix in self.avail_regions:
                    region_matrix.iloc[region_row, region_index] = 1
                    consecutive +=1
                    max_consecutive_col = consecutive
                else:
                    region_matrix.iloc[region_row, region_index] = 0
                    consecutive = 0
        row_indexes = self.find_most_ones(region_matrix)
        return row_indexes





import numpy as np
import matplotlib.pyplot as plt
from textwrap import wrap
import pandas as pd


class PrepareData:
    @staticmethod
    def check_input(sequencing_report, region_of_interest, method):
        assert isinstance(sequencing_report, pd.DataFrame)
        assert type(region_of_interest) == str
        assert region_of_interest in sequencing_report.columns.to_list()
        assert type(method) == str
        assert method in ["Shannon", "InverseSimpson"]

    @staticmethod
    def calc_simpson_index(clones):
        return 1 / (np.sum(clones**2))

    @staticmethod
    def calc_shannon_index(clones):
        return -np.sum(clones * np.log(clones))

    def cleaning(self, sequencing_report, region_of_interest, method):
        self.check_input(sequencing_report, region_of_interest, method)
        values = []
        unique_experiments = sequencing_report["Experiment"].unique().tolist()
        for experiment in unique_experiments:
            exp_spec = sequencing_report[sequencing_report["Experiment"] == experiment]
            aa_seqs = exp_spec[region_of_interest]
            clones = exp_spec["cloneFraction"]
            if method == "InverseSimpson":
                values.append(self.calc_simpson_index(clones))
            if method == "Shannon":
                values.append(self.calc_shannon_index(clones))
        return values, unique_experiments


class DiversityPlot:
    def __init__(
        self,
        sequencing_report,
        region_of_interest,
        ax=None,
        font_settings={},
        method="InverseSimpson",
    ):
        self.ax = ax
        self.method = method
        self.font_settings = font_settings
        values, unique_experiments = PrepareData().cleaning(
            sequencing_report, region_of_interest, method
        )
        if ax != None:
            self.ax = ax
            self.create_base_plot(values, unique_experiments)
            self.add_plot_addons(unique_experiments)
            self.set_title()

    def create_base_plot(
        self, values, unique_experiments, alpha=1, color="lightskyblue"
    ):
        self.ax.bar(
            x=unique_experiments,
            height=values,
            label="Diversity",
            color=color,
            alpha=alpha,
        )

    def add_plot_addons(self, unique_experiments):
        if self.method == "InverseSimpson":
            self.ax.set_ylabel("Inverse Simpson Index", **self.font_settings)
            plt.yscale("log")
        if self.method == "Shannon":
            self.ax.set_ylabel("Shannon Index", **self.font_settings)
        self.ax.set_xlabel("Sample", **self.font_settings)
        self.ax.set_xticklabels(
            labels=unique_experiments, rotation=45, ha="right", size=12
        )

    def set_title(self):
        if len(self.font_settings) != 0:
            original_fontsize = self.font_settings["fontsize"]
            self.font_settings["fontsize"] = 22
            if self.method == "InverseSimpson":
                title = "Diversity based on Inverse Simpson Index"
            if self.method == "Shannon":
                title = "Diversity based on Shannon Index"
            title = "\n".join(wrap(title, 40))
            self.ax.set_title(title, pad=12, **self.font_settings)
            self.font_settings["fontsize"] = original_fontsize



import os
import sys


def get_basename(files):
    basename = os.path.basename(files[0]).split(".")[0]
    if len(basename) > 22:
        basename = print(
            "Please change the name of your files so that they are shorter than 22 characters.\nThese names will later appear on the plots and they should be meaningful but also short enough.\nTherefore, the analysis has to be restarted."
        )
        sys.exit()
    return basename


class CreateCommand:
    def __init__(
        self,
        module_dir,
        path_to_mixcr,
        paired_end_sequencing,
        experiment,
        files,
        java_heap_size,
        threads=None,
    ):
        self.java_heap_size = java_heap_size
        self.module_dir = module_dir
        self.path_to_mixcr = path_to_mixcr
        self.threads = threads
        self.paired_end_sequencing = paired_end_sequencing
        self.files = files
        basename = get_basename(self.files)
        self.basename = basename
        self.result = os.path.join(self.module_dir, "temp", self.basename + ".vdjca")
        self.alignment_path = os.path.normpath(
            os.path.join(
                self.module_dir,
                "my_experiments",
                experiment,
                "alignment_reports",
                self.basename + "_AlignmentReport.txt",
            )
        )
        self.assembly_path = os.path.normpath(
            os.path.join(
                self.module_dir,
                "my_experiments",
                experiment,
                "assembly_reports",
                self.basename + "_AssemblyReport.txt",
            )
        )
        self.clones = os.path.join(
            self.module_dir,
            "my_experiments",
            experiment,
            "clones_result",
            self.basename + "_clones.clns",
        )
        self.clones_base = os.path.dirname(self.clones)
        self.table_tsv = os.path.join(
            module_dir,
            "my_experiments",
            experiment,
            "tables_mixcr",
            self.basename + ".tsv",
        )
        self.mixcr_plots_path = os.path.normpath(
            os.path.join(self.module_dir, "my_experiments", experiment, "mixcr_plots")
        )

    def read_aligned_reads(self):
        with open(self.alignment_path, "r") as f:
            alignment = f.readlines()
        for i in alignment:
            if "Successfully aligned reads" in i:
                quality = i.replace("Successfully aligned reads: ", "")
        align_quality = quality.split(" ")[0]
        percent_quality = quality.split(" ")[1]
        align_quality = int(align_quality)
        return align_quality, percent_quality

    def read_clones(self):
        with open(self.assembly_path, "r") as f:
            alignment = f.readlines()
        for i in alignment:
            if "Reads used in clonotypes, percent of total" in i:
                quality = i.replace("Reads used in clonotypes, percent of total: ", "")
        align_quality = quality.split(" ")[0]
        percent_quality = quality.split(" ")[1]
        assembly_quality = int(align_quality)
        return assembly_quality, percent_quality

    def get_free_memory_linux(self):
        with open("/proc/meminfo", "r") as f:
            lines = f.readlines()
        for line in lines:
            if line.startswith("MemAvailable"):
                return int(line.split()[1]) // 1024  # Convert from KB to MB

    def get_free_memory_windows(self):
        try:
            import psutil

            return psutil.virtual_memory().available // (
                1024**2
            )  # Convert from Bytes to MB
        except ImportError:
            print("Please install psutil library: pip install psutil")

    def get_free_memory_mac(self):
        cmd = "vm_stat | grep 'Pages free' | awk '{print $3}'"
        pages_free = int(os.popen(cmd).read().replace(".", ""))
        # Typically, a page in macOS is 4096 bytes
        return (pages_free * 4096) // (1024**2)  # Convert from Bytes to MB

    def create_parser(self):
        #   try:
        #      free_memory = self.get_free_memory()
        #      java_heap_size = int(free_memory / 4)  # Using half of the available RAM as an example
        #     print(f"25 % of currently available memory: {java_heap_size}\n")
        # except:
        #   print("automatic detection of memory failed. Processing continues with 1000MB RAM.")
        #  java_heap_size = 1000
        commands = []
        commands.extend(
            ["java", f"-Xms{self.java_heap_size}M", "-jar"]
        )  # enable change of para
        commands.extend([self.path_to_mixcr])
        return commands

    def prepare_align(self, method, add_args):
        align_commands = self.create_parser()
        align_commands.extend(["align"])
        align_commands.extend(["--preset", method])
        align_commands.extend(self.files)
        align_commands.extend([self.result])
        align_commands.extend(["--no-warnings"])
        align_commands.extend(["--force-overwrite"])
        if self.threads != None:
            align_commands.extend(["--threads", self.threads])
        align_commands.extend(["--report", self.alignment_path])
        for arg, value in add_args.items():
            if arg != "":
                align_commands.extend([f"--{arg}", value])
            else:
                pass
        return align_commands

    def prepare_assembly(self):

        assembly_commands = self.create_parser()
        assembly_commands.extend(["assemble"])
        # assembly_commands.extend(["-OseparateByC=true", "-OseparateByV=true","-OseparateByJ=true"])
        assembly_commands.extend([self.result])
        assembly_commands.extend([self.clones])
        assembly_commands.extend(["--force-overwrite"])
        assembly_commands.extend(["--report", self.assembly_path])

        return assembly_commands

    def prepare_clones(self, mixcr_chain):
        clones_commands = self.create_parser()
        clones_commands.extend(["exportClones"])
        clones_commands.extend([self.clones])
        clones_commands.extend(["--force-overwrite"])
        clones_commands.extend([f"--chains {mixcr_chain}"])
        clones_commands.extend([self.table_tsv])
        return clones_commands



from matplotlib_venn import venn2, venn3
from .matrices import morosita_horn_matrix
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

class PrepareData:
    @staticmethod
    def check_input(sequencing_report, samples):
        for sample in samples:
            assert sample in sequencing_report["Experiment"].to_list(), f"Please provide a valid sample name. The samples are: {sequencing_report['Experiment'].to_list()}"
        assert len(samples) == 2 or len(samples) == 3, "Please provide 2 or 3 samples"
        
    def tidy(self, sequencing_report:pd.DataFrame, samples:list, region_of_interest:str, ):
        self.check_input(sequencing_report, samples)
        PrepData = morosita_horn_matrix.PrepareData()
        matrix, unique_experiments = PrepData.cleaningPlot(samples, sequencing_report, region_string=region_of_interest, protein = True)
        readC_sample = sequencing_report.groupby("Experiment")["readCount"].sum() # the problem of this is that the an 80 % identity is not optimally visualized because a smaller circle (because of lower read count) can have issuesto fill 80 % of the bigger circle. 
        readC_sample = readC_sample.iloc[:len(samples)].values
        morosita_horn_base = matrix.loc[samples, samples].values
        morosita_horn_readC  = morosita_horn_base  # multiply identity with overall read count 
        morosita_horn_readC = np.around(morosita_horn_readC, 2)

        if morosita_horn_readC.shape[0] == 2:
            # Create a Venn diagram for 2 sets
            subsets=(morosita_horn_readC[0, 0], morosita_horn_readC[1, 1], morosita_horn_readC[1, 0])
        else:
            # Create a Venn diagram for 3 sets
            
            subsets=(morosita_horn_readC[0, 0], morosita_horn_readC[1, 1], morosita_horn_readC[1, 0], morosita_horn_readC[2, 2], morosita_horn_readC[2,0], morosita_horn_readC[2, 1], morosita_horn_readC[2, 1] * morosita_horn_readC[0, 1])
        
        return subsets, morosita_horn_base
    
class VennDiagram:
    def __init__(self, sequencing_report, samples, region_of_interest, font_settings = None, ax = None) :
        PrepData = PrepareData()
        subsets, morosita_horn_base = PrepData.tidy(sequencing_report, samples, region_of_interest)
        self.ax = ax
        if ax != None:
            self.create_venn(subsets, samples)

    def create_venn(self,subsets, samples):
            # Create a Venn diagram for 3 sets
        if len(samples) == 3:
            assert len(subsets) == 7, f"Yuu have {len(subsets)}"
            venn3(subsets, ax= self.ax, set_labels = samples)
        elif len(samples) == 2:
            venn2(subsets, ax= self.ax, set_labels = samples,)
            

            
    
    
morosita_horn = pd.read_csv(r"C:\Users\nilsh\my_projects\ExpoSeq\tmp_test\morosita_horn_test.csv")
sequencing_report = pd.read_csv(r"C:\Users\nilsh\my_projects\ExpoSeq\src\ExpoSeq\software_tests\test_files\test_show\sequencing_report.csv")
sequencing_report["cloneFraction"] = sequencing_report["readFraction"]
fig = plt.figure(1)
ax = fig.gca()
VennDiagram(sequencing_report, samples = ["GeneMind_1", "GeneMind_5", "GeneMind_6"], region_of_interest = "aaSeqCDR3", ax =ax)

plt.show()




import pandas as pd
import numpy as np
import editdistance
from scipy.spatial.distance import squareform
import warnings
from scipy.cluster.hierarchy import dendrogram, linkage
from textwrap import wrap
from matplotlib import pyplot as plt


class PrepareData:
    @staticmethod
    def check_input(samples, batch_size, max_cluster_dist, antigens):
        assert (
            type(samples) == list
        ), "You have to give a list as input for the sample"  # only one sample
        assert (
            type(max_cluster_dist) == int
        ), "You have to give an integer as input for the maximum levenshtein distance"
        assert (
            max_cluster_dist != 0
        ), "Plot does not make sense if you do not investigate at least SNPs"
        assert (
            type(batch_size) == int
        ), "You have to give an integer as input for the batch size"
        assert batch_size > 1
        assert antigens != None, "Please provide an antigen name"

    @staticmethod
    def tidy_data(
        sequencing_report,
        samples,
        batch_size,
        region_of_interest,
        antigens,
        binding_data,
    ):
        sample_report = sequencing_report[
            sequencing_report["Experiment"].isin(samples)
        ]  ## insert test if sample not found
        report = sample_report.groupby("Experiment").head(batch_size)
        report = report.drop_duplicates(subset=[region_of_interest])
        aa = report[region_of_interest]
        aa = pd.DataFrame(aa)
        pref_columns = antigens + [region_of_interest]
        for i in antigens:
            assert i in binding_data.columns.tolist(), f"{i} not in your binding data"
        assert region_of_interest in binding_data.columns.tolist()
        b_data = binding_data[pref_columns]
        mix = aa.merge(b_data, how="outer", on=region_of_interest)
        mix = mix.fillna(0)
        mix = mix.reset_index()
        aa_all = mix[region_of_interest]
        return mix, aa_all

    @staticmethod
    def create_distance_matrix(aa):
        num_sequences = len(aa)
        distance_matrix = np.zeros(
            (num_sequences, num_sequences)
        )  # initialize with zeros

        for i in range(num_sequences):
            for j in range(i + 1, num_sequences):
                lev_distance = editdistance.distance(aa[i], aa[j])
                distance_matrix[i, j] = lev_distance
                distance_matrix[j, i] = lev_distance  # The distance matrix is symmetric

        return distance_matrix

    @staticmethod
    def get_clustered_sequences(aa, max_cluster_dist):
        clustered_sequences = set()

        for i in range(len(aa)):
            for j in range(i + 1, len(aa)):
                if editdistance.distance(aa[i], aa[j]) < max_cluster_dist:
                    clustered_sequences.add(aa[i])
                    clustered_sequences.add(aa[j])

        return list(clustered_sequences)

    @staticmethod
    def label_bind_seqs(mix, region_string, aa_clustered, antigens):
        values = mix.loc[
            mix[region_string].isin(aa_clustered), antigens
        ]  # get the binding values for the sequences which were clustered
        # values = pd.DataFrame(values, columns = ["binding"])
        filtered_df = values[antigens][values[antigens] > 1].dropna(
            how="all"
        )  # filter out the binidng values which are 0
        indices = filtered_df.index.tolist()
        assert (
            len(indices) != 0
        ), "The cluster does not contain binding values. Please increase the lev. distance or the batch size."
        key_sequences = []
        key_values = []
        assert mix.columns.to_list() == ["index", "aaSeqCDR3"] + antigens
        for i in indices:
            key_sequences.append(mix.iloc[i][region_string])
            max_value = max(mix.iloc[i][column] for column in antigens)
            key_values.append(max_value)
        seq_val = pd.DataFrame([key_sequences, key_values]).T
        seq_val.columns = [
            "binding_seqs",
            "binding_values",
        ]  # seq_val object contains only the sequences from sanger - this is important for labelling
        seq_val = seq_val.sort_values(by="binding_values", ascending=False)
        seq_val.reset_index(inplace=True)
        binding_seqs = seq_val.binding_seqs.to_list()
        for label in range(len(aa_clustered)):
            label_str = aa_clustered[label]
            if label_str in binding_seqs:
                pos_label = binding_seqs.index(label_str)
                label_str = "(" + str(pos_label + 1) + ") " + label_str
                binding_seqs[pos_label] = label_str
            aa_clustered[label] = label_str
        return aa_clustered, binding_seqs, seq_val

    def process(
        self,
        sequencing_report,
        samples,
        batch_size,
        region_of_interest,
        antigens,
        binding_data,
        max_cluster_dist,
        sample_column_name="Experiment",
    ):
        for sample in samples:
            assert sample in list(
                sequencing_report[sample_column_name].unique()
            ), "The provided sample name does not exist"
        self.check_input(samples, batch_size, max_cluster_dist, antigens)
        mix, aa_all = self.tidy_data(
            sequencing_report,
            samples,
            batch_size,
            region_of_interest,
            antigens,
            binding_data,
        )
        aa_clustered = self.get_clustered_sequences(aa_all, max_cluster_dist)
        if len(aa_clustered) > 0:
            warnings.warn(
                "More than 30 sequences with Levenshtein distance < "
                + str(max_cluster_dist)
                + " found. The resulting plot could be disordered. To change that please reduce the batch size, max_cluster_dist or you can adjust the string size of the sequences on the y axis with: (1) ax = plot.ax and (2) ax.tick_params(axis='y', labelsize=your_desired_size)"
            )
        levenshtein_distance_matrix = self.create_distance_matrix(aa_clustered)
        condensed_matrix = squareform(
            levenshtein_distance_matrix, checks=False
        )  # linearize, now only 1 dim array
        linked = linkage(condensed_matrix, "single")
        aa_clustered, binding_seqs, seq_val = self.label_bind_seqs(
            mix, region_of_interest, aa_clustered, antigens
        )
        return linked, aa_clustered, binding_seqs, seq_val


class DendroBind:
    def __init__(
        self,
        sequencing_report,
        sample:list,
        region_of_interest,
        antigens,
        batch_size,
        max_cluster_dist,
        binding_data,
        ascending=True,
        fig=None,
        font_settings={},
    ) :
        linked, aa_clustered, binding_seqs, seq_val = PrepareData().process(
            sequencing_report,
            sample,
            batch_size,
            region_of_interest,
            antigens,
            binding_data,
            max_cluster_dist,
        )
        if fig != None:
            self.ax = fig.gca()
            self.fig2 = plt.figure(2)
            self.ax2 = self.fig2.gca()
            self.create_dendrogram(linked, aa_clustered)
            self.create_bar_plot(binding_seqs, seq_val, ascending)
        if font_settings != {}:
            self.set_dendrogram_labels(font_settings, sample)
            self.set_barplot_labels(font_settings)

    def create_dendrogram(self, linked, aa_clustered):
        dendrogram(
            linked,
            orientation="right",
            distance_sort="descending",
            show_leaf_counts=True,
            labels=aa_clustered,
            ax=self.ax,
        )

    def set_dendrogram_labels(
        self,
        font_settings:dict,
        sample:list,
    ):
        self.ax.set_xlabel("Levenshtein Distance", **font_settings)
        self.ax.set_ylabel("Sequences", **font_settings)
        title = "\n".join(
            wrap("Levenshtein Distance between sequences in " + " ".join(sample), 40)
        )

        self.ax.set_title(title, pad=12, **font_settings)

    @staticmethod
    def sort_binding_seqs(binding_seqs, seq_val, ascending=False):

        if ascending:

            zipped = list(zip(binding_seqs, seq_val.binding_values))
            # Sort by binding_values
            zipped_sorted = sorted(zipped, key=lambda x: x[1])

            # Unzip the sorted list
            binding_seqs_sorted, binding_values_sorted = zip(*zipped_sorted)
        else:
            binding_seqs_sorted = binding_seqs
            binding_values_sorted = seq_val.binding_values.to_list()
        return binding_seqs_sorted, binding_values_sorted

    def create_bar_plot(self, binding_seqs, seq_val, ascending):
        binding_seqs_sorted, binding_values_sorted = self.sort_binding_seqs(
            binding_seqs, seq_val, ascending
        )
        bars = self.ax2.barh(binding_seqs_sorted, binding_values_sorted)

    def set_barplot_labels(self, font_settings):
        self.ax2.set_ylabel("Sequences with binding data", **font_settings)
        self.ax2.set_xlabel("Binding Value", **font_settings)



from Bio.Align.Applications import ClustalwCommandline
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from ExpoSeq.design.msa_design import ExtendedSequenceManipulator
import os
import pandas as pd
import tkinter as tk
from tkinter import filedialog

class MSA:
    def __init__(self, fasta_path, out_path) -> None:
        clustalw_path = self.get_clustalw_path()
        self.run_msa(clustalw_path, fasta_path, out_path)
    
    def get_clustalw_path(self, Settings):
        global_params = Settings.read_global_vars()
        clustalw_path = global_params["clustalw_path"]
        if clustalw_path == '':
            print("Please choose the path to the clustalw exe with the filechooser. If you don't have it, you can download it from http://www.clustal.org/download/current/")
            clustalw_path = filedialog.askopenfilename()
            Settings.change_global_vars("clustalw_path", clustalw_path)

        return clustalw_path
    
    def run_msa(clustalw_path, seqs_path, outfile):         
        clustalw_cline = ClustalwCommandline(clustalw_path, infile=seqs_path, outfile=outfile)
        stdout, stderr = clustalw_cline()


metrics = [
    "Aboderin",
    "AbrahamLeo",
    "Argos",
    "Barley",
    "BlackMould",
    "BullBreese",
    "Casari",
    "Chothia",
    "Cid",
    "Cowan3.4",
    "Cowan7.5",
    "Eisenberg",
    "Engelman",
    "Fasman",
    "Fauchere",
    "Goldsack",
    "Guy",
    "HoppWoods",
    "Janin",
    "Jones",
    "Juretic",
    "Kidera",
    "Kuhn",
    "KyteDoolittle",
    "Levitt",
    "Manavalan",
    "Miyazawa",
    "Parker",
    "Ponnuswamy",
    "Prabhakaran",
    "Rao",
    "Rose",
    "Roseman",
    "Sweet",
    "Tanford",
    "Welling",
    "Wilson",
    "Wolfenden",
    "Zimmerman",
    "interfaceScale_pH2",
    "interfaceScale_pH8",
    "octanolScale_pH2",
    "octanolScale_pH8",
    "oiScale_pH2",
    "oiScale_pH8"
]


import peptides
from . import hytdrophobicity_metrics
from . import isoelectric_point_scales




class GetProteinProperty:

    def __init__(self, sequences):
        """This class allows you to return a dictionary with the sequences and the sequence property of interest

        Args:
            sequences (list): list of peptide sequences
        """
        self.sequences = sequences
        self.sequence_property_interest = {}
        self.attribute_funcs = {"aliphatic_index": self.get_aliphatic_index,
        "isoelectric_point": self.get_isoelectric_point,
        "hydrophobicity": self.get_hydrophobicity,
        "weight": self.get_weight,
        "mass_charge_ratio": self.mass_charge_ratio,
        "length": self.calc_len,
     }
        
    @staticmethod
    def calc_len(peptide_object):
        return len(peptide_object.sequence)
    
        
    @staticmethod
    def create_peptide_object(single_sequence):
        peptide = peptides.Peptide(single_sequence)
        return peptide
    
    @staticmethod
    def get_aliphatic_index(peptide_object):
        return peptide_object.aliphatic_index()
    
    @staticmethod
    def get_isoelectric_point(peptide_object, scale = "EMBOSS"):
        assert scale in isoelectric_point_scales.scale, f"Please choose on of the metrics from: {isoelectric_point_scales.scale}"
        return peptide_object.isoelectric_point(pKscale = scale)
    
    @staticmethod
    def get_hydrophobicity(peptide_object, scale = "KyteDoolittle"):
        assert scale in hytdrophobicity_metrics.metrics, f"Please choose on of the metrics from: {hytdrophobicity_metrics.metrics}"
        return peptide_object.hydrophobicity(scale = scale)
    
    @staticmethod
    def get_weight(peptide_object):
        return peptide_object.molecular_weight()
    
    @staticmethod
    def mass_charge_ratio(peptide_object):
        return peptide_object.mz()
    
    def calc_attribute(self, attribute, **kwargs):
        assert attribute in self.attribute_funcs.keys(), f"Please enter a value for attribute which is in {self.attribute_funcs.keys()}"
        for sequence in self.sequences:
            PepObj = self.create_peptide_object(single_sequence=sequence)
            func = self.attribute_funcs.get(attribute)
            attribute_value = func(PepObj, **kwargs)
            self.sequence_property_interest[sequence] = attribute_value
            
    def add_attributes_to_table(self, table, attribute):
        table[attribute] = list(self.sequence_property_interest.values())
        return table
      
    



import matplotlib.pyplot as plt
from . import plot_styler
import os
import matplotlib
import PyQt5

class MyFigure:
    def __init__(self, test, figure_style = "seaborn-v0_8-colorblind"):
        self.fig = plt.figure(1, figsize = (12, 10))
        self.ax = self.fig.gca()
        self.plot_type = "multi"
        self.figure_style = figure_style  
        self.test = test

    def check_fig(self, ):
        if not plt.fignum_exists(1):
            self.fig = plt.figure(1)

    def clear_fig(self):
        self.fig.clear()
        if self.plot_type == "single":
            self.ax = self.fig.gca()

    def update_plot(self):
        self.ax = self.fig.gca()
        self.use_style(self.figure_style)
        self.ax_visibility()
        self.style = plot_styler.PlotStyle(self.ax, self.plot_type)
        if matplotlib.get_backend() == "Qt5Agg" and self.test != True:
            figManager = plt.get_current_fig_manager()
            figManager.window.showMaximized()
        if self.test:
            self.tighten()
            plt.show(block = False)

        

    def tighten(self):
        plt.tight_layout()
    
    def set_backend(self, backend = "Qt5Agg"):
        if self.test != True:
            plt.switch_backend(backend)
        #return
        
    @staticmethod
    def use_style(style = "seaborn-v0_8-colorblind"):
        assert style in plt.style.available, f'The style you chose is not availale. Choose one of {plt.style.available}' 
        plt.style.use(style)
        
    def ax_visibility(self):
        self.ax.spines['right'].set_visible(False) 
        self.ax.spines['top'].set_visible(False)


def save_matrix(matrix, path = None):
    if path == None:
        while True:
            save_matrix = input("Do you want to save the generated data? (Y/n)")
            if save_matrix in ["Y", "y", "N", "n"]:
                break
            else:
                print("Please enter Y or n")
        if save_matrix.lower() in ["Y", "y"]:
            while True:
                filename_matrix = input("Enter a name for the file. The file will be saved locally in your IDE.")
                if not os.path.isfile(filename_matrix):
                    matrix.to_excel(path + ".xlsx")
                    break
                else:
                    print("This file already exists. Please choose another name.")
    else:
        matrix.to_excel(path)



class Test:
    def __init__(self) -> None:
        Fig = MyFigure()
        Fig.set_backend()
        test_ax = Fig.ax
        Fig.clear_fig()
        assert Fig.ax == test_ax
        Fig.ax.scatter(x = [1,2,3], y = [3,4,5])
        Fig.plot_type = "single"
        Fig.clear_fig()
        assert Fig.ax == Fig.fig.gca()
        
